from collections import defaultdict

import MinkowskiEngine as ME
import torch
import torch.nn as nn
import torch.nn.functional as F

from .experimental.bayes.decoder import MCDropoutDecoder
from .experimental.bayes.encoder import MCDropoutEncoder
from .experimental.bayes.evidential import EVDLoss
from .layer.cnn.act_norm import act_factory, norm_factory
from .layer.cnn.configuration import setup_cnn_configuration
from .layer.cnn.uresnet_layers import UResNet


class BayesianUResNet(torch.nn.Module):
    """
    UResNet with Uncertainty Quantification

    The backbone model consists of UResNet Encoder-Decoder format with
    standard residual layers for the shallow half and dropout residual layers
    for the deep half of the network.

    Configuration
    -------------
    mode: str
        string indicator for slight changes in network
        behavior/architecture. Supports three options:

            - standard: standard dropout segmentation network. This also
            includes MCDropout segnet, since training behavior is identical
            for both standard and mcdropout networks.
            - evd: Changes network into evidential segmentation network

    num_samples: int
        if used as MCDropout Segnet, the number of stochastic
        forward samples to be taken.

    num_classes: int
        number of segmentation classes (default: 5)

    """

    MODULES = []

    def __init__(self, cfg, name="mcdropout_uresnet"):
        super(BayesianUResNet, self).__init__()
        setup_cnn_configuration(self, cfg, "network_base")

        self.model_config = cfg.get(name, {})
        self.num_classes = self.model_config.get("num_classes", 5)
        self.num_samples = self.model_config.get("num_samples", 20)

        self.encoder = MCDropoutEncoder(cfg)
        self.decoder = MCDropoutDecoder(cfg)

        self.mode = self.model_config.get("mode", "standard")

        if "edl" in self.model_config.get("loss_fn", "cross_entropy"):
            self.classifier = nn.Sequential(
                ME.MinkowskiLinear(self.encoder.num_filters, self.num_classes),
                ME.MinkowskiSoftplus(),
            )
        else:
            self.classifier = ME.MinkowskiLinear(
                self.encoder.num_filters, self.num_classes
            )

    def mc_forward(self, input, num_samples=None):
        """
        Forwarding operation for MC Dropout segmentation network.

        Args:
            num_samples: number of stochastic forward samples to be taken
        """

        res = defaultdict(list)

        if num_samples is None:
            num_samples = self.num_samples

        for m in self.modules():
            if m.__class__.__name__ == "Dropout":
                m.train()

        for igpu, x in enumerate(input):

            num_voxels = x.shape[0]

            device = x.device

            x_sparse = ME.SparseTensor(
                coordinates=x[:, :4].int(), features=x[:, -1].view(-1, 1)
            )

            pvec = torch.zeros((num_voxels, self.num_classes)).to(device)
            logits = torch.zeros((num_voxels, self.num_classes)).to(device)

            for i in range(num_samples):
                res_encoder = self.encoder.encoder(x_sparse)
                decoderTensors = self.decoder(
                    res_encoder["finalTensor"], res_encoder["encoderTensors"]
                )
                feats = decoderTensors[-1]
                out = self.classifier(feats)
                logits += out.F
                pvec += F.softmax(out.F, dim=1)

            logits /= num_samples
            softmax_probs = pvec / num_samples

            res["softmax"].append(softmax_probs)
            res["segmentation"].append(logits)

        return res

    def evidential_forward(self, input):
        """
        Forwarding operation for evidential segmentation network.
        """
        out = defaultdict(list)
        for igpu, x in enumerate(input):
            x = ME.SparseTensor(
                coordinates=x[:, :4].int(), features=x[:, -1].view(-1, 1)
            )
            res_encoder = self.encoder.encoder(x)
            print([t.F.shape for t in res_encoder["encoderTensors"]])
            decoderTensors = self.decoder(
                res_encoder["finalTensor"], res_encoder["encoderTensors"]
            )
            feats = decoderTensors[-1]
            # For evidential models, logits correspond to collected evidence.
            logits = self.classifier(feats)
            ev = logits.F
            concentration = ev + 1.0
            S = torch.sum(concentration, dim=1, keepdim=True)
            uncertainty = self.num_classes / (S + 0.000001)
            out["segmentation"].append(ev)
            out["evidence"].append(ev)
            out["uncertainty"].append(uncertainty)
            out["concentration"].append(concentration)
            out["expected_probability"].append(concentration / S)
        return out

    def standard_forward(self, input):
        """
        Forwarding operation for standard dropout segmentation network.
        """
        out = defaultdict(list)
        for igpu, x in enumerate(input):
            x = ME.SparseTensor(
                coordinates=x[:, :4].int(), features=x[:, -1].view(-1, 1)
            )
            res_encoder = self.encoder.encoder(x)
            print([t.F.shape for t in res_encoder["encoderTensors"]])
            decoderTensors = self.decoder(
                res_encoder["finalTensor"], res_encoder["encoderTensors"]
            )
            feats = decoderTensors[-1]
            # For evidential models, logits correspond to collected evidence.
            logits = self.classifier(feats)
            out["segmentation"].append(logits.F)
        return out

    def forward(self, input):
        """ """

        if self.mode == "mc_dropout":
            return self.mc_forward(input)
        elif self.mode == "evidential":
            return self.evidential_forward(input)
        else:
            return self.standard_forward(input)


class DUQUResNet(torch.nn.Module):
    """
    Single Pass Deep Uncertainty Quantification Network
    Original Paper: https://arxiv.org/abs/2003.02037

    Implementation adapted from the DUQ main Github Repository:
    https://github.com/y0ast/deterministic-uncertainty-quantification

    Author: Joost van Amersfoort

    """

    MODULES = []

    def __init__(self, cfg, name="duq_uresnet"):
        super(DUQUResNet, self).__init__()
        setup_cnn_configuration(self, cfg, name)
        self.model_config = cfg.get(name, {})
        self.num_classes = self.model_config.get("num_classes", 5)
        self.num_samples = self.model_config.get("num_samples", 20)

        self.net = UResNet(cfg)

        self.gamma = self.model_config.get("gamma", 0.999)
        self.sigma = self.model_config.get("sigma", 0.3)

        self.embedding_dim = self.model_config.get("embedding_dim", 10)
        self.latent_size = self.model_config.get("latent_size", 32)

        self.w = nn.Parameter(
            torch.zeros(self.embedding_dim, self.num_classes, self.latent_size)
        )

        nn.init.kaiming_normal_(self.w, nonlinearity="relu")

        self.register_buffer("N", torch.ones(self.num_classes) * 20)
        self.register_buffer(
            "m", torch.normal(torch.zeros(self.embedding_dim, self.num_classes), 0.05)
        )

        self.m = self.m * self.N.unsqueeze(0)

    def embed(self, x):

        res = self.net(x)
        feats = res["decoderTensors"][-1]
        print(feats.F)
        out = torch.einsum("ij,mnj->imn", feats.F, self.w)
        return out

    def bilinear(self, z):
        embeddings = self.m / self.N.unsqueeze(0)

        diff = z - embeddings.unsqueeze(0)
        y_pred = (-(diff**2)).mean(1).div(2 * self.sigma**2).exp()

        return y_pred

    def forward(self, input):

        (point_cloud,) = input
        if self.training:
            point_cloud.requires_grad_(True)

        z = self.embed(point_cloud)
        y_pred = self.bilinear(z)

        res = {
            "score": [y_pred],
            "embedding": [z],
            "input": [point_cloud],
            "centroids": [
                self.m.detach().cpu().numpy() / self.N.detach().cpu().numpy()
            ],
        }

        self.z = z
        self.y_pred = y_pred

        return res

    def update_buffers(self):
        with torch.no_grad():
            # normalizing value per class, assumes y is one_hot encoded
            self.N = torch.max(
                self.gamma * self.N + (1 - self.gamma) * self.y_pred.sum(0),
                torch.ones_like(self.N),
            )
            # compute sum of embeddings on class by class basis
            features_sum = torch.einsum("ijk,ik->jk", self.z, self.y_pred)
            self.m = self.gamma * self.m + (1 - self.gamma) * features_sum


class SegmentationLoss(nn.Module):

    def __init__(self, cfg, name="mcdropout_uresnet"):
        super(SegmentationLoss, self).__init__()
        self.loss_config = cfg.get(name, {})
        self.loss_fn_name = self.loss_config.get("loss_fn", "edl_sumsq")
        self.loss_fn_args = self.loss_config.get("loss_fn_args", {})
        if "edl" in self.loss_fn_name:
            self.loss_fn = EVDLoss(self.loss_fn_name, **self.loss_fn_args)
        elif self.loss_fn_name == "cross_entropy":
            self.loss_fn = torch.nn.functional.cross_entropy
        else:
            raise ValueError(
                "Loss function {} not recognized".format(self.loss_fn_name)
            )
        self.one_hot = self.loss_config.get("one_hot", False)
        self.num_classes = self.loss_config.get("num_classes", 5)

    def forward(self, outputs, label, iteration=0, weight=None):
        """
        segmentation[0], label and weight are lists of size #gpus = batch_size.
        segmentation has as many elements as UResNet returns.
        label[0] has shape (N, dim + batch_id + 1)
        where N is #pts across minibatch_size events.
        """
        # TODO Add weighting
        logits = outputs["segmentation"]
        if "edl" in self.loss_fn_name:
            segmentation = [
                logits[0] + 1.0
            ]  # convert evidence to alpha concentration params.
        else:
            segmentation = logits
        device = segmentation[0].device
        assert len(segmentation) == len(label)
        # if weight is not None:
        #     assert len(data) == len(weight)
        batch_ids = [d[:, 0] for d in label]
        total_loss = 0
        total_acc = 0
        count = 0
        # Loop over GPUS
        for i in range(len(segmentation)):
            for b in batch_ids[i].unique():
                batch_index = batch_ids[i] == b
                event_segmentation = segmentation[i][batch_index]
                event_label = label[i][:, -1][batch_index]
                event_label = torch.squeeze(event_label, dim=-1).long()
                loss_label = event_label
                if self.one_hot:
                    loss_label = torch.eye(self.num_classes, device=device)[event_label]
                    loss_seg = self.loss_fn(event_segmentation, loss_label, t=iteration)
                else:
                    loss_seg = self.loss_fn(event_segmentation, loss_label)
                if weight is not None:
                    event_weight = weight[i][batch_index]
                    event_weight = torch.squeeze(event_weight, dim=-1)
                    total_loss += torch.mean(loss_seg * event_weight)
                else:
                    total_loss += torch.mean(loss_seg)
                # Accuracy
                predicted_labels = torch.argmax(event_segmentation, dim=-1)
                acc = (predicted_labels == event_label).sum().item() / float(
                    predicted_labels.nelement()
                )
                total_acc += acc
                count += 1

        return {"accuracy": total_acc / count, "loss": total_loss / count}


class DUQSegmentationLoss(nn.Module):

    def __init__(self, cfg, name="duq_uresnet"):
        super(DUQSegmentationLoss, self).__init__()
        self.xentropy = nn.BCELoss(reduction="none")
        self.num_classes = 5
        self.grad_w = cfg.get(name, {}).get("grad_w", 0.0)
        self.grad_penalty = cfg.get(name, {}).get("grad_penalty", True)

    @staticmethod
    def calc_gradient_penalty(x, y_pred):
        """
        Code From the DUQ main Github Repository:
        https://github.com/y0ast/deterministic-uncertainty-quantification

        Author: Joost van Amersfoort
        """
        gradients = torch.autograd.grad(
            outputs=y_pred,
            inputs=x,
            grad_outputs=torch.ones_like(y_pred),
            create_graph=True,
        )[0]

        gradients = gradients.flatten(start_dim=1)

        # L2 norm
        grad_norm = gradients.norm(2, dim=1)

        # Two sided penalty
        gradient_penalty = ((grad_norm - 1) ** 2).mean()

        # One sided penalty - down
        #     gradient_penalty = F.relu(grad_norm - 1).mean()

        return gradient_penalty

    def forward(self, out, type_labels):
        # print(type_labels)
        probas = out["score"][0]
        device = probas.device
        labels = type_labels[0][:, -1].long()
        labels_one_hot = torch.eye(self.num_classes)[labels].to(device=device)
        loss1 = self.xentropy(probas, labels_one_hot)
        pred = torch.argmax(probas, dim=1)

        # Comptue gradient penalty
        loss2 = 0
        if self.grad_penalty:
            loss2 = self.calc_gradient_penalty(out["input"][0], probas)

        loss1 = loss1.sum(dim=1).mean()
        loss = loss1 + self.grad_w * loss2

        accuracy = float(torch.sum(pred == labels)) / float(labels.shape[0])

        res = {
            "loss": loss,
            "loss_embedding": float(loss1),
            "loss_grad_penalty": float(loss2),
            "accuracy": accuracy,
        }

        print(res)

        acc_types = {}
        for c in labels.unique():
            mask = labels == c
            acc_types["accuracy_{}".format(int(c))] = float(
                torch.sum(pred[mask] == labels[mask])
            ) / float(torch.sum(mask))
        return res



"""Defines CNN encoder backbones for image feature extraction."""

import MinkowskiEngine as ME
import torch

from spine.utils.globals import COORD_COLS, VALUE_COL

from .uresnet_layers import UResNetEncoder

__all__ = ["SparseResidualEncoder"]


class SparseResidualEncoder(UResNetEncoder):
    """Encoder for sparse tensor feature extraction."""

    def __init__(self, coord_conv=False, pool_mode="avg", feature_size=512, **cfg):
        """Initializes the sparse residual CNN encoder.

        Passes most of the configuration along to the underlying sparse
        residual CNN encoder defined in :class:`UResNetEncoder`.

        Parameters
        ----------
        coord_conv : bool, default False
            Whether to include normalized coordinates in the input features
        pool_mode : str, default 'avg'
            Final pooling operation mode ('avg', 'max' or 'conv'
        feature_size : int, default 512
            Number of features produced after the final pooling
        **cfg : dict, optional
            Configuration to pass along to the sparse residual encoder
        """
        # Initialize the parent class
        super().__init__(cfg)

        # Store attributes
        self.coord_conv = coord_conv

        # Initialize the final pooling layer
        assert self.spatial_size is not None, (
            "Must specify `spatial_size` to know how many pooling stages "
            "are needed to get to a one dimensional vector."
        )
        final_tensor_shape = self.spatial_size // (2 ** (self.depth - 1))

        if pool_mode == "avg":
            # Average pooling
            self.pool = ME.MinkowskiGlobalAvgPooling()

        if pool_mode == "sum":
            # Sum pooling
            self.pool = ME.MinkowskiGlobalSumPooling()

        elif pool_mode == "max":
            # Max pooling
            self.pool = ME.MinkowskiGlobalMaxPooling()

        elif pool_mode == "conv":
            # Strided convolution
            self.pool = torch.nn.Sequential(
                ME.MinkowskiConvolution(
                    in_channels=self.num_planes[-1],
                    out_channels=self.num_planes[-1],
                    kernel_size=final_tensor_shape,
                    stride=final_tensor_shape,
                    dimension=self.dim,
                ),
                ME.MinkowskiGlobalPooling(),
            )

        else:
            raise ValueError(
                f"Pooling mode not recognized: {self.pool_mode}. Must be "
                "one of 'avg', 'sum', 'max' or 'conv'"
            )

        # Initialize the final linear layer
        self.feature_size = feature_size
        self.linear = ME.MinkowskiLinear(self.num_planes[-1], self.feature_size)

    def forward(self, data):
        """Pass one batch of data through the CNN encoder.

        Parameters
        ----------
        data : torch.Tensor
             (N, 1 + D + N_f) Batch of data

        Returns
        -------
        torch.Tensor
            (B) Batch of features, one per batch ID
        """
        # Build an input feature tensor
        coords = data[:, :VALUE_COL]
        features = data[:, VALUE_COL].view(-1, 1)

        # If requested, append the normalized coordinates to the feature tensor
        if self.coord_conv:
            normalized_coords = data[:, COORD_COLS] / self.spatial_size
            features = torch.cat([normalized_coords, features], dim=1)

        # Build a sparse tensor
        x = ME.SparseTensor(coordinates=coords.int(), features=features)

        # Pass through the CNN encoder
        output = super().forward(x)
        final_tensor = output["final_tensor"]

        # Pool the last layer
        z = self.pool(final_tensor)

        # Put it through a linear layer
        latent = self.linear(z)

        return latent.F



"""Construct an analysis script module class from its name."""

from spine.utils.factory import instantiate, module_dict

from . import calib, diag, metric, script

# Build a dictionary of available calibration modules
ANA_DICT = {}
for module in [calib, diag, metric, script]:
    ANA_DICT.update(**module_dict(module))


def ana_script_factory(name, cfg, overwrite=None, log_dir=None, prefix=None):
    """Instantiates an analyzer module from a configuration dictionary.

    Parameters
    ----------
    name : str
        Name of the analyzer module
    cfg : dict
        Analysis script module configuration
    parent_path : str
        Path to the parent directory of the main analysis configuration. This
        allows for the use of relative paths in the analyzers.
    overwrite : bool, optional
        If `True`, overwrite the CSV logs if they already exist
    log_dir : str, optional
        Output CSV file directory (shared with driver log)
    prefix : str, optional
        Input file prefix. If requested, it will be used to prefix
        all the output CSV files.

    Returns
    -------
    object
         Initialized analyzer object
    """
    # Provide the name to the configuration
    cfg["name"] = name

    # Instantiate the analysis script module
    if overwrite is not None:
        return instantiate(
            ANA_DICT, cfg, overwrite=overwrite, log_dir=log_dir, prefix=prefix
        )
    else:
        return instantiate(ANA_DICT, cfg, log_dir=log_dir, prefix=prefix)



"""Whole-image classification/regression tasks."""

import torch
import torch.nn as nn

from spine.data import TensorBatch

from .layer.factories import encoder_factory, loss_fn_factory

__all__ = ["ImageClassifier", "ClusterImageClassifier", "ImageClassLoss"]


class ImageClassifier(nn.Module):
    """Whole-image classification model.

    This model uses various encoder declinations to classifier an entire
    image as belonging to a certain class.

    .. code-block:: yaml
        model:
          name: image_class
          modules:
            classifier:
              # Image classifier configuration
            classifier_loss:
              # Image classifier loss configuration
    """

    MODULES = ["classifier", "classifier_loss"]

    def __init__(self, classifier, classifier_loss=None):
        """Initialize the particle image classifier.

        Parameters
        ----------
        classifier : dict
            Image classifier configuration
        classifier_loss : dict, optional
            Image classifier loss configuration
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the model
        self.process_model_config(**classifier)

    def process_model_config(self, num_classes, **encoder):
        """Initialize the underlying encoder and the final layer.

        Parameters
        ----------
        num_classes : int
            Number of classes that each image can be sorted as
        **encoder : dict
            Encoder configuration
        """
        # Initialize the encoder
        self.encoder = encoder_factory(encoder)
        """
        self.encoder_type = model_cfg.get('encoder_type', 'standard')
        if self.encoder_type == 'dropout':
            self.encoder = MCDropoutEncoder(cfg)
        elif self.encoder_type == 'standard':
            self.encoder = SparseResidualEncoder(cfg)
        elif self.encoder_type == 'pointnet':
            self.encoder = PointNetEncoder(cfg)
        else:
            raise ValueError('Unrecognized encoder type: {}'.format(self.encoder_type))
        """

        # Initialize the final layer
        self.final_layer = nn.Linear(self.encoder.feature_size, num_classes)

    def forward(self, data):
        """Run a batch of data through the forward function.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
            - N is the the total number of voxels in the image
            - 1 is the batch ID
            - D is the number of dimensions in the input image
            - N_f is the number of features per voxel
        """
        # Pass the input through the encoder
        out = self.encoder(data.tensor)

        # Pass the features through the final layer
        logits = self.final_layer(out)

        # Store the output as a tensor batch
        logits = TensorBatch(logits, counts=[1] * data.batch_size)

        # Return
        return {"logits": logits}


class ImageClassLoss(nn.Module):
    """Image classication loss."""

    def __init__(self, classifier, classifier_loss):
        """Intialize the image classification loss.

        Parameters
        ----------
        classifier : dict
            Image classifier configuration
        classifier_loss : dict, optional
            Image classifier loss configuration
        """
        # Initialize the parent class
        super().__init__()

        # Extract the number of logits output by the classifier
        self.process_model_config(**classifier)

        # Initialize the loss function
        self.process_loss_config(**classifier_loss)

    def process_model_config(self, num_classes, **encoder):
        """Initialize the underlying encoder and the final layer.

        Parameters
        ----------
        num_classes : int
            Number of classes that each image can be sorted as
        **encoder : dict
            Encoder configuration
        """
        self.num_classes = num_classes

    def process_loss_config(self, loss="ce", balance_loss=False, weights=None):
        """Initialize the loss function

        Parameters
        ----------
        loss : str, default 'ce'
            Name of the loss function to apply
        balance_loss : bool, default False
            Whether to weight the loss to account for class imbalance
        weights : list, optional
            (C) One weight value per class
        """
        # Initialize basic parameters
        self.balance_loss = balance_loss
        self.weights = weights

        # Sanity check
        assert (
            weights is None or not balance_loss
        ), "Do not provide weights if they are to be computed on the fly."

        # Set the loss
        self.loss_fn = loss_fn_factory(loss, functional=True)

    def forward(self, labels, logits, **kwargs):
        """Applies the image classification loss to a batch of data.

        Parameters
        ----------
        labels : List[int]
            (B) List of image labels, one per entry in the batch
        logits : TensorBatch
            (B, C) Tensor of predicted logits, one per entry in the batch
        **kwargs : dict, optional
            Other labels/outputs of the model which are not relevant here

        Returns
        -------
        loss : torch.Tensor
            Value of the loss
        accuracy : float
            Value of the image-wise classification accuracy
        """
        # Cast the labels to a long torch Tensor
        labels = torch.tensor(labels, dtype=torch.long, device=logits.device)

        # Create a mask for valid images (-1 indicates an invalid class ID)
        valid_mask = labels > -1

        # Check that the labels and the output tensor size are compatible
        num_classes = logits.shape[1]
        class_mask = labels < num_classes
        if torch.any(~class_mask):
            warn(
                "There are class labels with a value larger than the "
                f"size of the output logit vector ({num_classes})."
            )

        valid_mask &= class_mask

        # Apply the valid mask and convert the labels to a torch.Tensor
        valid_index = torch.where(valid_mask)[0]
        labels = labels[valid_index]
        logits = logits.tensor[valid_index]

        # Compute the loss. Balance classes if requested
        if self.balance_loss:
            self.weights = get_class_weights(labels, num_classes=num_classes)

        loss = self.loss_fn(logits, labels, weight=self.weights, reduction="sum")
        if len(valid_index):
            loss /= len(valid_index)

        # Compute accuracy of assignment (fraction of correctly assigned images)
        acc = 1.0
        acc_class = [1.0] * num_classes
        if len(valid_index):
            preds = torch.argmax(logits, dim=1)
            acc = float(torch.sum(preds == labels))
            acc /= len(valid_index)
            for c in range(num_classes):
                index = torch.where(labels == c)[0]
                if len(index):
                    acc_class[c] = float(torch.sum(preds[index] == c)) / len(index)

        # Prepare and return result
        result = {"loss": loss, "accuracy": acc}

        for c in range(num_classes):
            result[f"accuracy_class_{c}"] = acc_class[c]

        return result


class ClusterImageClassifier(nn.Module):

    pass



import numpy as np
import torch
import torch.nn as nn

from spine.utils.factory import instantiate


def calibrator_dict():

    models = {"temperature_scaling": TemperatureScaling}

    return models


def calibrator_loss_dict():

    losses = {"CalibrationNLLLoss": CalibrationNLLLoss}

    return losses


def calibrator_factory(cfg):
    """Instantiates a calibrator from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        Calibrator configuration

    Returns
    -------
    object
        Instantiated calibrator
    """
    calibrators = calibrator_dict()
    return instantiate(calibrators, cfg)


def calibrator_loss_factory(cfg):
    """Instantiates a calibrator loss from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        Calibrator loss configuration

    Returns
    -------
    object
        Instantiated calibrator loss
    """
    calibrator_losses = calibrator_loss_dict()
    return instantiate(calibrator_losses, cfg)


class CalibrationNLLLoss(nn.Module):

    def __init__(self, logit_name, **kwargs):
        super(CalibrationNLLLoss, self).__init__()
        self.loss_fn = nn.CrossEntropyLoss(**kwargs)
        self.logit_name = logit_name

    def forward(self, res, labels):
        logits = res[self.logit_name][0]

        targets = labels[0][:, 0].to(dtype=torch.long)

        loss = self.loss_fn(logits, targets)
        pred = torch.argmax(logits, dim=1)

        accuracy = float(torch.sum(pred == targets)) / float(targets.shape[0])

        out = {"loss": loss, "accuracy": accuracy, "T": float(res["T"][0])}

        acc_types = {}
        for c in targets.unique():
            mask = targets == c
            acc_types["accuracy_{}".format(int(c))] = float(
                torch.sum(pred[mask] == targets[mask])
            ) / float(torch.sum(mask))
        return out


class PostHocCalibrationModel(nn.Module):
    """
    Base Class for Post-Hoc Uncertainty Calibration Methods

    Post-hoc calibration methods are trained on a validation set
    after its client model's (the model to be calibrated) training
    has converged in the training set.

    The client model must be freezed so that the parameters does not
    change over the course of training the calibration model.

    Also, it is desirable that the calibration model preserves the
    ordering of the logit predictions, so that the accuracy of the
    client model is completely unchanged after calibration (isotonicity)
    """

    def __init__(self, model, calibration_cfg):
        super(PostHocCalibrationModel, self).__init__()
        self.model = model
        self.model.eval()
        self.cfg = calibration_cfg
        if "logit_name" not in self.cfg:
            raise KeyError("Calibrator needs output logit specifier name")
        self.logit_name = self.cfg["logit_name"]
        self.calibration_params = None

    def train(self, mode: bool = True):
        # Override .train() to always set client model to eval mode.
        super().train(mode)
        self.model.eval()

    def forward(self, input):
        raise NotImplementedError


class TemperatureScaling(PostHocCalibrationModel):

    def __init__(self, model, calibration_cfg):
        super(TemperatureScaling, self).__init__(model, calibration_cfg)
        self.calibration_params = nn.Parameter(torch.ones(1) * 1.5)

    def temperature_scale(self, logits):
        T = self.calibration_params.expand(logits.shape)
        return logits / T

    def forward(self, input):

        res = self.model(input)
        logits = res[self.logit_name][0]
        res[self.logit_name] = [self.temperature_scale(logits)]

        res["T"] = [self.calibration_params.detach()]

        return res



"""SPINE driver class.

Takes care of everything in one centralized place:
- Data loading
- ML model and loss forward pass
- Batch unwrapping
- Representation building
- Post-processing
- Analysis script execution
- Writing output to file
"""

import os
import subprocess as sc
import time
from datetime import datetime

import numpy as np
import psutil
import yaml

from .ana import AnaManager
from .banner import ascii_logo
from .construct import BuildManager
from .geo import GeoManager
from .io import reader_factory, writer_factory
from .io.core.write.csv import CSVWriter
from .math import seed as numba_seed
from .model import ModelManager
from .post import PostManager
from .utils.conditional import TORCH_AVAILABLE
from .utils.logger import logger
from .utils.stopwatch import StopwatchManager
from .utils.torch import runtime
from .utils.torch.devices import set_visible_devices
from .utils.unwrap import Unwrapper
from .version import __version__

__all__ = ["Driver"]


class Driver:
    """Central SPINE driver.

    Processes global configuration and runs the appropriate modules:
      1. Load data
      2. Run the model forward (including loss) and backward (if training)
      3. Unwrap batched data
      4. Build representations
      5. Run post-processing
      6. Run analysis scripts
      7. Write to file

    It takes a configuration dictionary of the form:

    .. code-block:: yaml

        base:
          <Base driver configuration>
        geo:
          <Geometry configuration>
        io:
          <Input/output configuration>
        model:
          <Model architecture>
        build:
          <Rules as to how to build reconstructed object representations>
        post:
          <Post-processors>
        ana:
          <Analysis scripts>
    """

    def __init__(self, cfg, rank=None):
        """Initializes the class attributes.

        Parameters
        ----------
        cfg : dict
            Global configuration dictionary
        rank : int, optional
            Rank of the GPU. If not specified, the model will be run on CPU if
            `world_size` is 0 and GPU is `world_size` is > 0.
        """
        # Initialize the timers and the configuration dictionary
        self.watch = StopwatchManager()
        self.watch.initialize("iteration")

        # Process the full configuration dictionary and store it
        base, io, geo, model, build, post, ana = self.process_config(**cfg, rank=rank)

        # Initialize the base driver configuration parameters
        train = self.initialize_base(**base, rank=rank)

        # Initialize the detector geometry singleton once and for all modules
        if geo is not None:
            GeoManager.initialize_or_get(**geo)

        # Initialize the input/output
        self.initialize_io(**io)

        # Initialize the ML model
        self.model = None
        if model is not None:
            assert (
                self.loader is not None
            ), "The model can only be used in conjunction with a loader."
            self.watch.initialize("model")

            # Check if PyTorch is available for model functionality
            if not TORCH_AVAILABLE:
                raise ImportError(
                    "PyTorch is required for model functionality. "
                    "Install with: pip install spine-ml[model]"
                )

            self.model = ModelManager(
                **model,
                train=train,
                dtype=self.dtype,
                rank=self.rank,
                distributed=self.distributed,
            )

        else:
            assert (
                train is None
            ), "Received a train block but there is no model to train."

        self.builder = None
        # Initialize the data representation builder
        if build is not None:
            assert (
                self.model is None or self.unwrap
            ), "Must unwrap the model output to build representations."
            assert (
                self.model is None or self.model.to_numpy
            ), "Must cast model output to numpy to build representations."
            self.watch.initialize("build")
            self.builder = BuildManager(**build)

        # Initialize the post-processors
        self.post = None
        if post is not None:
            assert (
                self.model is None or self.unwrap
            ), "Must unwrap the model output to run post-processors."
            self.watch.initialize("post")
            self.post = PostManager(
                post,
                post_list=self.post_list,
                parent_path=self.parent_path,
            )

        # Initialize the analysis scripts
        self.ana = None
        if ana is not None:
            assert (
                self.model is None or self.unwrap
            ), "Must unwrap the model output to run analysis scripts."
            self.watch.initialize("ana")
            self.ana = AnaManager(ana, log_dir=self.log_dir, prefix=self.log_prefix)

    def process_config(
        self,
        io=None,
        base=None,
        geo=None,
        model=None,
        build=None,
        post=None,
        ana=None,
        rank=None,
    ):
        """Reads the configuration and dumps it to the logger.

        Parameters
        ----------
        io : dict
            I/O configuration dictionary
        base : dict, optional
            Base driver configuration dictionary
        geo : dict, optional
            Geometry configuration dictionary
        model : dict, optional
            Model configuration dictionary
        build : dict, optional
            Representation building configuration dictionary
        post : dict, optional
            Post-processor configuration dictionary
        ana : dict, optional
            Analysis script configuration dictionary
        rank : int, optional
            Rank of the GPU. The model will be run on CPU if `world_size` is not
            specified or 0 and on GPU is `world_size` is > 0.

        Returns
        -------
        dict
            Processed configuration
        """
        # If there is no base configuration, make it empty (will use defaults)
        if base is None:
            base = {}

        # Set the verbosity of the logger
        verbosity = base.get("verbosity", "info")
        logger.setLevel(verbosity.upper())

        # Set GPUs visible to CUDA (function handles torch availability)
        base["world_size"] = set_visible_devices(**base)

        # If the seed is not set for the sampler, randomize it. This is done
        # here to keep a record of the seeds provided to the samplers
        assert io is not None, "The `io` block is always required."
        if "loader" in io:
            if "sampler" in io["loader"]:
                if isinstance(io["loader"]["sampler"], str):
                    io["loader"]["sampler"] = {"name": io["loader"]["sampler"]}

                if (
                    "seed" not in io["loader"]["sampler"]
                    or io["loader"]["sampler"]["seed"] < 0
                ):
                    io["loader"]["sampler"]["seed"] = int(time.time())

        # If the seed is not set for the training/inference process, randomize it
        if "seed" not in base or base["seed"] < 0:
            base["seed"] = int(time.time())
        else:
            assert isinstance(
                base["seed"], int
            ), f"The driver seed must be an integer, got: {base['seed']}"

        # Rebuild global configuration dictionary
        self.cfg = {"base": base, "io": io}
        if geo is not None:
            self.cfg["geo"] = geo
        if model is not None:
            self.cfg["model"] = model
        if build is not None:
            self.cfg["build"] = build
        if post is not None:
            self.cfg["post"] = post
        if ana is not None:
            self.cfg["ana"] = ana

        # Log information for the main process only
        if rank is None or rank < 1:
            # Log package logo
            logger.info("\n%s", ascii_logo)

            # Log environment information
            logger.info("Release version: %s\n", __version__)

            visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES", None)
            logger.info("$CUDA_VISIBLE_DEVICES=%s\n", visible_devices)

            system_info = sc.getstatusoutput("uname -a")[1]
            logger.info("Configuration processed at: %s\n", system_info)

            # Log configuration
            logger.info(yaml.dump(self.cfg, default_flow_style=None, sort_keys=False))

        # Return updated configuration
        return base, io, geo, model, build, post, ana

    def initialize_base(
        self,
        seed,
        dtype="float32",
        world_size=None,
        gpus=None,
        log_dir="logs",
        prefix_log=False,
        overwrite_log=False,
        parent_path=None,
        iterations=None,
        epochs=None,
        unwrap=False,
        rank=None,
        log_step=1,
        distributed=False,
        torch_sharing_strategy=None,
        split_output=False,
        train=None,
        verbosity="info",
    ):
        """Initialize the base driver parameters.

        Parameters
        ----------
        seed : int
            Random number generator seed
        dtype : str, default 'float32'
            Data type of the model parameters and input data
        world_size : int, optional
            Number of GPUs to use in the underlying model
        gpus : List[int], optional
            List of indexes of GPUs to expose to the model
        log_dir : str, default 'logs'
            Path to the directory where the logs will be written to
        prefix_log : bool, default False
            If True, use the input file name to prefix the log name
        overwrite_log : bool, default False
            If True, overwrite log even if it already exists
        parent_path : str, optional
            Path to the parent directory of the analysis configuration file
        iterations : int, optional
            Number of entries/batches to process (-1 means all entries)
        epochs : int, optional
            Number of times to iterate over the full dataset
        unwrap : bool, default False
            Wheather to unwrap batched data (only relevant when using loader)
        rank : int, optional
            Rank of the GPU in the multi-GPU training process
        log_step : int, default 1
            Number of iterations before the logging is called (1: every step)
        distributed : bool, default False
            If `True`, this process is distributed among multiple processes
        train : dict, optional
            Training configuration dictionary
        split_output : bool, default False
            Split the output of the process into one file per input file
        verbosity : int, default 'info'
            Verbosity level to pass to the `logging` module. Pick one of
            'debug', 'info', 'warning', 'error', 'critical'.

        Returns
        -------
        dict
            Training configuration
        rank
            Updated rank
        """
        # Set up the seed
        np.random.seed(seed)
        numba_seed(seed)
        runtime.manual_seed(seed)

        # Set up the device the model will run on
        if rank is None and world_size > 0:
            assert (
                world_size < 2
            ), "Must not request > 1 GPU without specifying a GPU rank."
            rank = 0

        self.rank = rank
        self.world_size = world_size
        self.main_process = rank is None or rank == 0

        # Check on the distributed process
        assert self.rank is None or self.rank < world_size, (
            f"The GPU rank index of this driver ({rank}) is too large "
            f"for the number of GPUs available ({world_size})."
        )

        self.distributed = distributed
        if not distributed and world_size > 1:
            self.distributed = True

        # Store general parameters
        self.dtype = dtype
        self.log_dir = log_dir
        self.prefix_log = prefix_log
        self.overwrite_log = overwrite_log
        self.parent_path = parent_path
        self.iterations = iterations
        self.epochs = epochs
        self.unwrap = unwrap
        self.seed = seed
        self.log_step = log_step
        self.split_output = split_output

        return train

    def initialize_io(self, loader=None, reader=None, writer=None):
        """Initializes the input/output scripts.

        Parameters
        ----------
        loader : dict, optional
            PyTorch DataLoader configuration dictionary
        reader : dict, optional
            Reader configuration dictionary
        writer : dict, optional
            Writer configuration dictionary
        """
        # Make sure that we have either a data loader or a reader, not both
        assert (loader is not None) ^ (
            reader is not None
        ), "Must provide either a loader or a reader configuration."

        # Initialize the data loader/reader
        self.loader = None
        self.loader_iter = None
        self.counter = None
        self.unwrapper = None
        if loader is not None:
            # Initialize the torch data loader - requires PyTorch
            if not TORCH_AVAILABLE:
                raise ImportError(
                    "PyTorch is required for loader functionality. "
                    "Install with: pip install spine-ml[model]"
                )

            # Import loader_factory only when PyTorch is available
            from .io import loader_factory

            self.watch.initialize("load")
            self.loader = loader_factory(
                **loader,
                rank=self.rank,
                dtype=self.dtype,
                world_size=self.world_size,
                distributed=self.distributed,
            )

            self.iter_per_epoch = len(self.loader)
            self.reader = self.loader.dataset.reader

            # If requested, initialize the unwrapper
            if self.unwrap:
                self.watch.initialize("unwrap")
                self.unwrapper = Unwrapper()

            # If working from LArCV files, no post-processor was yet run
            self.post_list = ()

        else:
            # Initialize the reader
            self.watch.initialize("read")
            self.reader = reader_factory(reader)
            self.iter_per_epoch = len(self.reader)

            # Fetch the list of previously run post-processors
            # TODO: this only works with two runs in a row, not 3 and above
            self.post_list = None
            if self.reader.cfg is not None and "post" in self.reader.cfg:
                self.post_list = tuple(self.reader.cfg["post"])

        # Fetch an appropriate common prefix for all input files
        self.log_prefix, self.output_prefix = self.get_prefixes(
            self.reader.file_paths, self.split_output
        )

        # Initialize the data writer, if provided
        self.writer = None
        if writer is not None:
            assert (
                self.loader is None or self.unwrap
            ), "Must unwrap the model output to write it to file."
            self.watch.initialize("write")
            self.writer = writer_factory(
                writer, prefix=self.output_prefix, split=self.split_output
            )

        # Harmonize the iterations and epochs parameters
        assert (self.iterations is None) or (
            self.epochs is None
        ), "Must not specify both `iterations` or `epochs` parameters."
        if self.iterations is not None:
            if self.iterations < 0:
                self.iterations = self.iter_per_epoch
            self.epochs = 1.0
        elif self.epochs is not None:
            self.iterations = int(self.epochs * self.iter_per_epoch)

    @staticmethod
    def get_prefixes(file_paths, split_output):
        """Builds an appropriate output prefix based on the list of input files.

        Parameters
        ----------
        file_paths : List[str]
            List of input file paths
        split_output : bool
            Split the output of the process into one file per input file

        Returns
        -------
        Union[str, List[str]]
            Shared input summary string to be used to prefix outputs
        """
        # Fetch file base names (ignore where they live)
        file_names = [os.path.splitext(os.path.basename(f))[0] for f in file_paths]

        # Get the shared prefix of all files in the list
        prefix = os.path.commonprefix(file_names)

        # If there is only one file, done
        if len(file_names) == 1:
            if not split_output:
                return prefix, prefix
            else:
                return prefix, [prefix]

        # Otherwise, assemble log name from input file names
        sep = "--"
        log_prefix = ""
        if len(prefix):
            log_prefix += prefix

        # Get the shared suffix of all files in the list
        file_names_f = [f[::-1] for f in file_names]
        suffix = os.path.commonprefix(file_names_f)[::-1]
        if prefix == suffix:
            suffix = ""

        # Pad the center of the log name with compnents which are not shared
        first = file_names[0][len(prefix) : len(file_names[0]) - len(suffix)]
        if len(first):
            if len(log_prefix):
                log_prefix += sep
            log_prefix += first

        skip_count = len(file_names) - 2
        if len(file_names) > 2:
            if len(log_prefix):
                log_prefix += sep
            log_prefix += f"{skip_count}"

        last = file_names[-1][len(prefix) : len(file_names[-1]) - len(suffix)]
        if len(last):
            if len(log_prefix):
                log_prefix += sep
            log_prefix += last

        # Add the shared suffix
        if len(suffix):
            log_prefix += f"--{suffix}"

        # Truncate file names that are too long
        max_length = 150
        if len(log_prefix) > max_length:
            log_prefix = log_prefix[: max_length - 3] + "---"

        # Always provide a single prefix for the log, adapt output prefix
        if not split_output:
            return log_prefix, log_prefix
        else:
            return log_prefix, file_names

    def initialize_log(self):
        """Initialize the output log for this driver process."""
        # Make a directory if it does not exist
        if self.log_dir and not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir, exist_ok=True)

        # Determine the log name, initialize it
        if self.builder is not None or self.model is None:
            # If running the driver more than a model, give a generic name
            log_name = f"spine_log.csv"
        else:
            # If running the driver within a training/validation process
            # (model only), follow a specific pattern of log names.
            start_iteration = self.model.start_iteration
            prefix = "train" if self.model.train else "inference"
            suffix = "" if not self.model.distributed else f"_proc{self.rank}"
            log_name = f"{prefix}{suffix}_log-{start_iteration:07d}.csv"

        # If requested, prefix the log name with the input file name
        if self.prefix_log:
            log_name = f"{self.log_prefix}_{log_name}"

        # Initialize the log
        log_path = os.path.join(self.log_dir, log_name)
        self.logger = CSVWriter(log_path, overwrite=self.overwrite_log)

    def __len__(self):
        """Returns the number of events in the underlying reader object.

        Returns
        -------
        int
            Number of elements in the underlying loader/reader.
        """
        return len(self.reader)

    def __iter__(self):
        """Resets the counter and returns itself.

        Returns
        -------
        object
            The Driver itself
        """
        # If a loader is used, reinitialize it. Otherwise set an entry counter
        if self.loader is not None:
            self.loader_iter = iter(self.loader)
        else:
            self.counter = 0

        return self

    def __next__(self):
        """Defines how to process the next entry in the iterator.

        Returns
        -------
        Union[dict, List[dict]]
            Either one combined data dictionary, or one per entry in the batch
        """
        # If there are more iterations to go through, return data
        if self.loader is not None:
            return self.process()
        else:
            self.counter = 0 if self.counter is None else self.counter
            if self.counter < len(self):
                data = self.process(self.counter)
                self.counter += 1

                return data

            raise StopIteration

    def run(self):
        """Loop over the requested number of iterations, process them."""
        # To run the loop, must know how many times it must be done
        assert (
            self.iterations is not None
        ), "Must specify either `iterations` or `epochs` parameters."

        # Initialize the output log
        self.initialize_log()

        # Get the iteration start (if model exists)
        start_iteration = 0
        if self.model is not None and self.model.train:
            start_iteration = self.model.start_iteration

        # Loop and process each iteration
        for iteration in range(start_iteration, self.iterations):
            # When switching to a new epoch, reset the loader iterator
            if self.loader is not None and (
                self.loader_iter is None or iteration % self.iter_per_epoch == 0
            ):
                if self.distributed:
                    epoch_cnt = iteration // self.iter_per_epoch
                    self.loader.sampler.set_epoch(epoch_cnt)
                self.loader_iter = iter(self.loader)

            # Update the epoch counter, record the execution date/time
            epoch = (iteration + 1) / self.iter_per_epoch
            tstamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

            # Process one batch/entry of data
            entry = iteration if self.loader is None else None
            data = self.process(entry=entry, iteration=iteration)

            # Log the output
            self.log(data, tstamp, iteration, epoch)

            # Release the memory for the next iteration
            data = None

    def process(self, entry=None, run=None, subrun=None, event=None, iteration=None):
        """Process one entry or a batch of entries.

        Run single step of main SPINE driver. This includes data loading,
        model forwarding, data structure building, post-processing
        and appending desired information to each row of output csv files.

        Parameters
        ----------
        entry : int, optional
            Entry number to load
        run : int, optional
            Run number to load
        subrun : int, optional
            Subrun number to load
        event : int, optional
            Event number to load
        iteration : int, optional
            Iteration number. Only needed to train models and/or to apply
            time-dependant model losses, no-op otherwise

        Returns
        -------
        Union[dict, List[dict]]
            Either one combined data dictionary, or one per entry in the batch
        """
        # 0. Make sure there is no watch running, start the iteration timer
        for watch in self.watch.values():
            if watch.running or watch.paused:
                self.watch.reset()
                break

        self.watch.start("iteration")

        # 1. Load data
        data = self.load(entry, run, subrun, event)

        # 2. Pass data through the model
        if self.model is not None:
            self.watch.start("model")
            result = self.model(data, iteration=iteration)
            data.update(**result)
            self.watch.stop("model")
            self.watch.update(self.model.watch, "model")

        # 3. Unwrap
        if self.unwrapper is not None:
            self.watch.start("unwrap")
            data = self.unwrapper(data)
            self.watch.stop("unwrap")

        # 4. Build representations
        if self.builder is not None:
            self.watch.start("build")
            self.builder(data)
            self.watch.stop("build")

        # 5. Run post-processing, if requested
        if self.post is not None:
            self.watch.start("post")
            self.post(data)
            self.watch.stop("post")
            self.watch.update(self.post.watch, "post")

        # 6. Run scripts, if requested
        if self.ana is not None:
            self.watch.start("ana")
            self.ana(data)
            self.watch.stop("ana")
            self.watch.update(self.ana.watch, "ana")

        # 7. Write output to file, if requested
        if self.writer is not None:
            self.watch.start("write")
            self.writer(data, self.cfg)
            self.watch.stop("write")

        # Stop the iteration timer
        self.watch.stop("iteration")

        # Return
        return data

    def load(self, entry=None, run=None, subrun=None, event=None):
        """Loads one batch/entry to process.

        If the model is run on the fly, the data is batched. Otherwise,
        a single entry is loaded at this stage.

        Parameters
        ----------
        entry : int, optional
            Entry number, only valid with reader
        run : int, optional
            Run number, only valid with reader
        subrun : int, optional
            Subrun number to load
        event : int, optional
            Event number, only valid with reader

        Returns
        -------
        data: dict
            Data dictionary containing the input
        """
        # Dispatch to the appropriate loader
        if self.loader is not None:
            # Can only load batches sequentially, not by index
            assert entry is None and run is None and subrun is None and event is None, (
                "When calling the loader, there is no way to request a "
                "specific entry or run/subrun/event triplet."
            )

            # Initialize the loader, if necessary
            if self.loader_iter is None:
                self.loader_iter = iter(self.loader)

            # Load the next batch
            self.watch.start("load")
            data = next(self.loader_iter)
            self.watch.stop("load")

        else:
            # Must provide either entry number or both run and event numbers
            assert (entry is not None) or (
                run is not None and subrun is not None and event is not None
            ), (
                "Provide either the entry number or the run, subrun "
                "and event number to read."
            )

            # Read an entry
            self.watch.start("read")
            if entry is not None:
                data = self.reader.get(entry)
            else:
                data = self.reader.get_run_event(run, subrun, event)
            self.watch.stop("read")

        return data

    def apply_filter(
        self,
        n_entry=None,
        n_skip=None,
        entry_list=None,
        skip_entry_list=None,
        run_event_list=None,
        skip_run_event_list=None,
    ):
        """Restrict the list of entries.

        Parameters
        ----------
        n_entry : int, optional
            Maximum number of entries to load
        n_skip : int, optional
            Number of entries to skip at the beginning
        entry_list : list, optional
            List of integer entry IDs to add to the index
        skip_entry_list : list, optional
            List of integer entry IDs to skip from the index
        run_event_list: list((int, int, int)), optional
            List of (run, subrun, event) triplets to add to the index
        skip_run_event_list: list((int, int, int)), optional
            List of (run, subrun, event) triplets to skip from the index
        """
        # Simply change the underlying entry list
        self.reader.process_entry_list(
            n_entry,
            n_skip,
            entry_list,
            skip_entry_list,
            run_event_list,
            skip_run_event_list,
        )

        # Reset the iterator
        self.loader_iter = None

    def log(self, data, tstamp, iteration, epoch=None):
        """Log relevant information to CSV files and stdout.

        Parameters
        ----------
        data : dict
            Dictionary of data products to extract scalars from
        tstamp : str
            Time when this iteration was run
        iteration : int
            Iteration counter
        epoch : float
            Progress in the training process in number of epochs
        """
        # Fetch the first entry in the batch
        first_entry = data["index"]
        if isinstance(first_entry, list):
            first_entry = first_entry[0]

        # Fetch the basics
        log_dict = {"iter": iteration, "epoch": epoch, "first_entry": first_entry}

        # Fetch the memory usage (in GB)
        log_dict["cpu_mem"] = psutil.virtual_memory().used / 1.0e9
        log_dict["cpu_mem_perc"] = psutil.virtual_memory().percent
        log_dict["gpu_mem"], log_dict["gpu_mem_perc"] = 0.0, 0.0
        if runtime.cuda_is_available():
            gpu_total = runtime.cuda_mem_info()[-1] / 1.0e9
            log_dict["gpu_mem"] = runtime.cuda_max_memory_allocated() / 1.0e9
            log_dict["gpu_mem_perc"] = 100 * log_dict["gpu_mem"] / gpu_total

        # Fetch the times
        suff = "_time"
        for key, watch in self.watch.items():
            time_iter, time_sum = watch.time, watch.time_sum
            log_dict[f"{key}{suff}"] = time_iter.wall
            log_dict[f"{key}{suff}_cpu"] = time_iter.cpu
            log_dict[f"{key}{suff}_sum"] = time_sum.wall
            log_dict[f"{key}{suff}_sum_cpu"] = time_sum.cpu

        # Fetch all the scalar outputs and append them to a dictionary
        for key in data:
            if np.isscalar(data[key]):
                log_dict[key] = data[key]
            elif runtime.is_tensor(data[key]) and data[key].dim() == 0:
                log_dict[key] = data[key].item()

        # Record
        self.logger.append(log_dict)

        # If requested, log out basics of the training/inference process
        log = ((iteration + 1) % self.log_step) == 0
        if log:
            # Dump general information
            proc = (
                "train" if self.model is not None and self.model.train else "inference"
            )
            device = "GPU" if self.rank is not None else "CPU"
            keys = [f"Time ({proc})", f"{device} memory", "Loss", "Accuracy"]
            widths = [20, 20, 9, 9]
            if self.distributed:
                keys = ["Rank"] + keys
                widths = [5] + widths
            if self.main_process:
                header = "  | " + "| ".join(
                    [f"{keys[i]:<{widths[i]}}" for i in range(len(keys))]
                )
                separator = "  |" + "+".join(["-" * (w + 1) for w in widths])
                msg = f"Iter. {iteration} (epoch {epoch:.3f}) @ {tstamp}\n"
                msg += header + "|\n"
                msg += separator + "|"
                logger.info(msg)
            if self.distributed:
                runtime.distributed_barrier()

            # Dump information pertaining to a specific process
            t_iter = self.watch.time("iteration").wall
            t_net = 0.0
            if self.model is not None:
                t_net = self.watch.time("model").wall

            if self.rank is not None:
                mem, mem_perc = log_dict["gpu_mem"], log_dict["gpu_mem_perc"]
            else:
                mem, mem_perc = log_dict["cpu_mem"], log_dict["cpu_mem_perc"]

            acc = data.get("accuracy", -1.0)
            loss = data.get("loss", -1.0)

            values = [
                f"{t_iter:0.2f} s ({100*t_net/t_iter:0.2f} %)",
                f"{mem:0.2f} GB ({mem_perc:0.2f} %)",
                f"{loss:0.3f}",
                f"{acc:0.3f}",
            ]
            if self.distributed:
                values = [f"{self.rank}"] + values

            msg = "  | " + "| ".join(
                [f"{values[i]:<{widths[i]}}" for i in range(len(keys))]
            )
            msg += "|"
            logger.info(msg)

            # Start new line once only
            if self.distributed:
                runtime.distributed_barrier()
            if self.main_process:
                logger.info("")



"""Module that handles conditional imports for optional packages.

Currently wraps the following packages:
- ROOT: only needed when parsing ROOT-format data
- larcv: only needed when reading larcv-format data in parsers
- torch: only needed when running ML models and training
- MinkowskiEngine: only needed when running sparse CNNs
"""

import os
from warnings import warn

__all__ = [
    "ROOT",
    "larcv",
    "torch",
    "ME",
    "MF",
    "ROOT_AVAILABLE",
    "LARCV_AVAILABLE",
    "TORCH_AVAILABLE",
    "ME_AVAILABLE",
]


# If ROOT is available, load it
try:
    import ROOT

    ROOT_AVAILABLE = True
except ModuleNotFoundError:
    warn("ROOT could not be found, cannot parse LArCV data.")
    ROOT = None
    ROOT_AVAILABLE = False


# If LArCV is available, load it
try:
    from larcv import larcv

    LARCV_AVAILABLE = True
except ModuleNotFoundError:
    warn("larcv could not be found, cannot parse LArCV data.")
    larcv = None
    LARCV_AVAILABLE = False


# If torch is available, load it
try:
    import torch

    TORCH_AVAILABLE = True
except ModuleNotFoundError:
    warn("PyTorch could not be found, ML functionality disabled.")

    # Create a mock torch module for type annotations
    class MockTorch:
        """Mock torch module that provides attributes for type annotations."""

        class Tensor:
            """Mock torch.Tensor for type annotations."""

            pass

        def __getattr__(self, name):
            """Return None for any other attributes."""
            return None

    torch = MockTorch()
    TORCH_AVAILABLE = False


# If MinkowskiEngine is available, load it with the right number of threads
try:
    if os.environ.get("OMP_NUM_THREADS") is None:
        os.environ["OMP_NUM_THREADS"] = "16"
    import MinkowskiEngine as ME
    import MinkowskiFunctional as MF

    ME_AVAILABLE = True
except ModuleNotFoundError:
    warn("MinkowskiEngine could not be found, cannot run sparse CNNs.")
    ME = None
    MF = None
    ME_AVAILABLE = False



"""Main configuration loading functions for SPINE.

This module provides the primary entry points for loading SPINE configurations:
- load_config(): Load from a YAML string
- load_config_file(): Load from a file path
- _load_config_recursive(): Internal recursive loader with include support
"""

import os
import warnings
from typing import Any, Dict, List, Optional, Tuple

import yaml

from .api import META_DESCRIPTION, META_KEY, META_LIST_APPEND, META_STRICT, META_VERSION
from .errors import ConfigCycleError, ConfigIncludeError
from .loader import ConfigLoader, resolve_config_path
from .meta import check_compatibility, extract_metadata
from .operations import (
    _apply_overrides_and_removals,
    apply_collection_operation,
    deep_merge,
    extract_includes_and_overrides,
    parse_value,
    set_nested_value,
)

__all__ = ["load_config", "load_config_file"]


def _load_config_recursive(
    cfg_path: Optional[str] = None,
    config_string: Optional[str] = None,
    root_dir: Optional[str] = None,
    include_stack: Optional[List[str]] = None,
    compatibility_checks: Optional[List[Tuple[Dict, Dict, str]]] = None,
) -> Tuple[Dict[str, Any], Dict[str, Any], List[str], Dict[str, Any]]:
    """Recursively load config with cycle detection.

    Parameters
    ----------
    cfg_path : Optional[str]
        Path to configuration file (mutually exclusive with config_string)
    config_string : Optional[str]
        YAML configuration string (mutually exclusive with cfg_path)
    root_dir : Optional[str]
        Root directory for resolving relative include paths.
        Required when using config_string with includes.
        Defaults to directory of cfg_path when loading from file.
    include_stack : Optional[List[str]]
        Stack of currently-loading files (for cycle detection)
    compatibility_checks : Optional[List[Tuple[Dict, Dict, str]]]
        List to accumulate (parent_meta, included_meta, path) for deferred checking

    Returns
    -------
    Tuple[Dict[str, Any], Dict[str, Any], List[str], Dict[str, Any]]
        (config content, override directives, removal directives, metadata)

    Raises
    ------
    ConfigCycleError
        If circular include detected
    ConfigIncludeError
        If included file not found
    ValueError
        If both or neither cfg_path and config_string are provided
    """
    # Validate inputs
    if (cfg_path is None) == (config_string is None):
        raise ValueError("Must provide exactly one of cfg_path or config_string")

    # Determine the identifier for cycle detection and root directory
    if cfg_path is not None:
        cfg_path = os.path.abspath(cfg_path)
        identifier = cfg_path
        if root_dir is None:
            root_dir = os.path.dirname(cfg_path)
    else:
        # For string configs, use a pseudo-identifier
        identifier = "<string>"
        if root_dir is None:
            root_dir = os.getcwd()

    # Cycle detection
    if include_stack is None:
        include_stack = []
    if compatibility_checks is None:
        compatibility_checks = []

    if identifier in include_stack and cfg_path is not None:
        cycle = include_stack + [identifier]
        raise ConfigCycleError(cycle)

    include_stack = include_stack + [identifier]

    # Create a custom loader class with the specified root_dir
    class CustomConfigLoader(ConfigLoader):
        def __init__(self, stream):
            super().__init__(stream, root_dir)

    # Load YAML
    try:
        if cfg_path is not None:
            with open(cfg_path, "r", encoding="utf-8") as f:
                main_config = yaml.load(f, Loader=CustomConfigLoader)
        else:
            assert config_string is not None  # For the linter's sake
            main_config = yaml.load(config_string, Loader=CustomConfigLoader)
    except FileNotFoundError as exc:
        raise ConfigIncludeError(f"Configuration file not found: {cfg_path}") from exc
    except Exception as exc:
        source = cfg_path if cfg_path else "<string>"
        raise ConfigIncludeError(f"Error loading {source}: {exc}") from exc

    if main_config is None:
        return {}, {}, [], {}

    # Extract metadata
    metadata = extract_metadata(main_config, cfg_path if cfg_path else "<string>")
    strict = metadata[META_STRICT]
    list_append_mode = metadata[META_LIST_APPEND]

    # Extract directives
    includes, overrides, removals, cleaned_config = extract_includes_and_overrides(
        main_config
    )

    # Remove __meta__ from cleaned config
    if META_KEY in cleaned_config:
        del cleaned_config[META_KEY]

    config = {}

    # Process includes
    for include_file in includes:
        # Resolve include path with SPINE_CONFIG_PATH support
        include_path = resolve_config_path(include_file, root_dir)

        # Recursively load
        (
            included_config,
            included_overrides,
            included_removals,
            included_meta,
        ) = _load_config_recursive(
            cfg_path=include_path,
            root_dir=None,
            include_stack=include_stack,
            compatibility_checks=compatibility_checks,
        )

        # Warn if included file has no metadata (but keep the metadata that was extracted)
        if not included_meta.get(META_VERSION) and not included_meta.get(
            META_DESCRIPTION
        ):
            # File likely has no __meta__ block (only defaults)
            file_name = os.path.basename(include_path)
            warnings.warn(
                f"Included file '{file_name}' has no __meta__ block. "
                f"Consider adding metadata for better configuration management.",
                stacklevel=2,
            )

        # Defer compatibility check until all includes loaded
        compatibility_checks.append((metadata, included_meta, include_path))

        # Merge included config
        config = deep_merge(config, included_config)

        # Merge component versions from included metadata into parent
        # This allows subsequent includes to check against accumulated components
        if "components" in included_meta:
            if "components" not in metadata:
                metadata["components"] = {}
            metadata["components"].update(included_meta["components"])
        elif META_VERSION in included_meta:
            # If included file has version but no components, infer component name from file path
            # e.g., base/base_240719.yaml -> component "base" with version "240719"
            # This allows configs without explicit components to still participate in version checking
            include_dir = os.path.basename(os.path.dirname(include_path))
            if include_dir and include_dir not in ("", "."):
                if "components" not in metadata:
                    metadata["components"] = {}
                metadata["components"][include_dir] = included_meta[META_VERSION]

        # Apply included overrides (use included file's strict/list_append settings)
        included_strict = included_meta.get(META_STRICT, strict)
        included_list_append = included_meta.get(META_LIST_APPEND, list_append_mode)

        config, unapplied = _apply_overrides_and_removals(
            config,
            included_overrides,
            included_removals,
            included_strict,
            included_list_append,
        )

        # Propagate unapplied overrides
        if unapplied:
            overrides = {**unapplied, **overrides}

    # Merge main config content
    if cleaned_config:
        config = deep_merge(config, cleaned_config)

    return config, overrides, removals, metadata


def load_config(config_str: str, root_dir: Optional[str] = None) -> Dict[str, Any]:
    """Load a SPINE configuration from a YAML string.

    Similar to yaml.safe_load(), but with SPINE's advanced features:
    - Hierarchical includes with cycle detection
    - Metadata via __meta__ blocks
    - Override semantics with dot-notation
    - Collection operations (list append/remove, dict key removal)
    - Configurable strict modes (warn/error)

    See module docstring for full configuration language spec.

    Parameters
    ----------
    config_str : str
        YAML configuration string
    root_dir : Optional[str]
        Root directory for resolving relative include paths.
        Also used as the base for SPINE_CONFIG_PATH searches.
        If not provided, defaults to current working directory.
        Required if config contains __include__ directives with relative paths.

    Returns
    -------
    Dict[str, Any]
        Loaded and merged configuration

    Raises
    ------
    ConfigCycleError
        If circular include detected
    ConfigIncludeError
        If included file not found or can't be loaded
    ConfigPathError
        If removal/operation targets non-existent path (when strict="error")
    ConfigTypeError
        If operation applied to wrong type
    ConfigOperationError
        If invalid operation specified

    Examples
    --------
    Simple string config:

    >>> config_str = \"\"\"
    ... io:
    ...   reader:
    ...     batch_size: 32
    ... \"\"\"
    >>> config = load_config(config_str)
    >>> print(config['io']['reader']['batch_size'])
    32

    String config with includes (requires root_dir):

    >>> config_str = \"\"\"
    ... include: base.yaml
    ... model:
    ...   name: resnet
    ... \"\"\"
    >>> config = load_config(config_str, root_dir="/path/to/configs")

    For loading from files, use load_config_file():

    >>> config = load_config_file("config.yaml")

    Or equivalently:

    >>> with open("config.yaml") as f:
    ...     config = load_config(f.read())

    See Also
    --------
    load_config_file : Load configuration from a file path
    """
    # Load recursively, accumulating compatibility checks
    compatibility_checks = []
    config, overrides, removals, metadata = _load_config_recursive(
        config_string=config_str,
        root_dir=root_dir,
        compatibility_checks=compatibility_checks,
    )

    # Now that all includes are loaded, check all compatibility requirements
    for parent_meta, included_meta, include_path in compatibility_checks:
        check_compatibility(parent_meta, included_meta, include_path)

    # Get strict and list_append settings from top-level metadata
    # (these are always present, set by extract_metadata with defaults)
    strict = metadata[META_STRICT]
    list_append_mode = metadata[META_LIST_APPEND]

    # Apply top-level overrides
    # Note: these include both explicit top-level overrides and propagated ones from nested files
    # Use strict mode from top-level metadata
    for key_path, value in overrides.items():
        parsed_value = parse_value(value)

        if key_path.endswith("+") or key_path.endswith("-"):
            # Collection operations - use strict mode from metadata
            base_key = key_path[:-1]
            operation = key_path[-1]
            config = apply_collection_operation(
                config, base_key, parsed_value, operation, strict, list_append_mode
            )
        else:
            # Regular override - silently skip if parent doesn't exist
            config, _ = set_nested_value(
                config, key_path, parsed_value, only_if_exists=True
            )

    # Apply top-level removals
    for key_path in removals:
        config, _ = set_nested_value(config, key_path, None, delete=True, strict=strict)

    # Remove __meta__ from final config
    if META_KEY in config:
        del config[META_KEY]

    return config


def load_config_file(cfg_path: str) -> Dict[str, Any]:
    """Load a SPINE configuration from a file.

    Convenience function that reads a configuration file and passes it to load_config.
    The file's directory is automatically used as root_dir for include resolution.

    Parameters
    ----------
    cfg_path : str
        Path to configuration file

    Returns
    -------
    Dict[str, Any]
        Loaded and merged configuration

    Raises
    ------
    ConfigCycleError
        If circular include detected
    ConfigIncludeError
        If included file not found or can't be loaded
    ConfigPathError
        If removal/operation targets non-existent path (when strict="error")
    ConfigTypeError
        If operation applied to wrong type
    ConfigOperationError
        If invalid operation specified

    Examples
    --------
    >>> config = load_config_file("config.yaml")
    >>> print(config['io']['reader']['batch_size'])
    32

    See Also
    --------
    load_config : Load config from a YAML string
    """
    # Load recursively, accumulating compatibility checks
    compatibility_checks = []
    cfg_path = os.path.abspath(cfg_path)
    root_dir = os.path.dirname(cfg_path)

    config, overrides, removals, metadata = _load_config_recursive(
        cfg_path=cfg_path, root_dir=root_dir, compatibility_checks=compatibility_checks
    )

    # Now that all includes are loaded, check all compatibility requirements
    for parent_meta, included_meta, include_path in compatibility_checks:
        check_compatibility(parent_meta, included_meta, include_path)

    # Get strict and list_append settings from top-level metadata
    # (these are always present, set by extract_metadata with defaults)
    strict = metadata[META_STRICT]
    list_append_mode = metadata[META_LIST_APPEND]

    # Apply top-level overrides
    # Note: these include both explicit top-level overrides and propagated ones from nested files
    # Use strict mode from top-level metadata
    for key_path, value in overrides.items():
        parsed_value = parse_value(value)

        if key_path.endswith("+") or key_path.endswith("-"):
            # Collection operations - use strict mode from metadata
            base_key = key_path[:-1]
            operation = key_path[-1]
            config = apply_collection_operation(
                config, base_key, parsed_value, operation, strict, list_append_mode
            )
        else:
            # Regular override - silently skip if parent doesn't exist
            config, _ = set_nested_value(
                config, key_path, parsed_value, only_if_exists=True
            )

    # Apply top-level removals
    for key_path in removals:
        config, _ = set_nested_value(config, key_path, None, delete=True, strict=strict)

    # Remove __meta__ from final config
    if META_KEY in config:
        del config[META_KEY]

    return config



"""Post-processor in charge of finding matches between charge and light."""

import numpy as np

from spine.data.out.base import OutBase
from spine.geo import GeoManager
from spine.post.base import PostBase
from spine.utils.optical import FlashMerger

from .barycenter import BarycenterFlashMatcher
from .likelihood import LikelihoodFlashMatcher

__all__ = ["FlashMatchProcessor"]


class FlashMatchProcessor(PostBase):
    """Associates TPC interactions with optical flashes."""

    # Name of the post-processor (as specified in the configuration)
    name = "flash_match"

    # Alternative allowed names of the post-processor
    aliases = ("run_flash_matching",)

    # Whether this post-processor needs to know where the configuration lives
    need_parent_path = True

    # Whether the post-processor needs geometry information
    need_geometry = True

    def __init__(
        self,
        flash_key,
        volume,
        ref_volume_id=None,
        method="likelihood",
        time_contained=False,
        max_cathode_offset=None,
        run_mode="reco",
        truth_point_mode="points",
        truth_dep_mode="depositions",
        parent_path=None,
        merge=None,
        update_flashes=False,
        **kwargs,
    ):
        """Initialize the flash matching algorithm.

        Parameters
        ----------
        flash_key : str
            Flash data product name. In most cases, this is unambiguous, unless
            there are multiple types of segregated optical detectors
        volume : str
            Physical volume corresponding to each flash ('module' or 'tpc')
        ref_volume_id : str, optional
            If specified, the flash matching expects all interactions/flashes
            to live into a specific optical volume. Must shift everything.
        method : str, default 'likelihood'
            Flash matching method (one of 'likelihood' or 'barycenter')
        time_contained : bool, default False
            If `True`, only match interactions which are time contained
        max_cathode_offset : float, optional
            If specified, only match cathode-crossing interactions which are
            offset from the cathode less than this threshold
        parent_path : str, optional
            Path to the parent directory of the main analysis configuration.
            This allows for the use of relative paths in the post-processors.
        merge : dict, optional
            Flash merging configuration
        update_flashes : bool, default False
            If `True` and merging flashes, replaces the original list of
            flashes in place with the list of merged flashes
        **kwargs : dict
            Keyword arguments to pass to specific flash matching algorithms
        """
        # Initialize the parent class
        super().__init__(
            "interaction",
            run_mode,
            truth_point_mode,
            truth_dep_mode,
            parent_path=parent_path,
        )

        # Make sure the flash data product is available, store
        self.flash_key = flash_key
        self.update_keys({flash_key: True})

        # Get the volume within which each flash is confined
        assert volume in (
            "tpc",
            "module",
        ), "The `volume` must be one of 'tpc' or 'module'."
        self.volume = volume
        self.ref_volume_id = ref_volume_id

        # Store the timing checks to be performed
        self.time_contained = time_contained
        self.max_cathode_offset = max_cathode_offset
        if self.time_contained:
            self.update_upstream("time_containment")
        if self.max_cathode_offset is not None:
            self.update_upstream("cathode_crosser")

        # Fetch the detector geometry
        self.geo = GeoManager.get_instance()

        # Initialize the flash matching algorithm
        if method == "barycenter":
            self.matcher = BarycenterFlashMatcher(**kwargs)

        elif method == "likelihood":
            self.matcher = LikelihoodFlashMatcher(
                detector=self.geo.name.lower(), parent_path=self.parent_path, **kwargs
            )

        else:
            raise ValueError(f"Flash matching method not recognized: {method}")

        # Initialize the flash merging class, if needed
        self.merger = None
        if merge is not None:
            self.merger = FlashMerger(**merge)
        self.update_flashes = update_flashes

    def process(self, data):
        """Find [interaction, flash] pairs.

        Parameters
        ----------
        data : dict
            Dictionary of data products

        Notes
        -----
        This post-processor modifies the list of `interaction` objects
        in-place by filling the following attributes:
        - interaction.is_flash_matched: (bool)
               Indicator for whether the given interaction has a flash match
        - interaction.flash_ids: np.ndarray
               The flash IDs in the flash list
        - interaction.flash_volume_ids: np.ndarray
               The flash optical volume IDs in the flash list
        - interaction.flash_times: np.ndarray
               The flash time(s) in microseconds
        - interaction.flash_scores: np.ndarray
               The flash scores(s) (larger is better)
        - interaction.flash_total_pe: float
               Total number of PEs associated with the matched flash(es)
        - interaction.flash_hypo_pe: float, optional
               Total number of PEss associated with the hypothesis flash
        """
        # Fetch the optical flashes
        flashes = data[self.flash_key]

        # Merge flashes based on timing, if requested
        orig_ids = np.arange(len(flashes))
        if self.merger is not None:
            flashes, orig_ids = self.merger(flashes)

        # Loop over the optical volumes, run flash matching
        volume_ids = np.asarray([f.volume_id for f in flashes])
        for k in self.interaction_keys:
            # Fetch interactions, nothing to do if there are not any
            interactions = data[k]
            if not len(interactions):
                continue

            # Make sure the interaction coordinates are expressed in cm
            self.check_units(interactions[0])

            # Clear previous flash matching information
            for inter in interactions:
                inter.reset_flash_match(typed=False)

            # Loop over the optical volumes
            for volume_id in np.unique(volume_ids):
                # Get the list of flashes associated with this optical volume
                flashes_v = []
                for flash in flashes:
                    if flash.volume_id == volume_id:
                        flashes_v.append(flash)

                # Crop interactions to only include depositions in the optical volume
                interactions_v = []
                for inter in interactions:
                    # If requested, skip interactions which are not time contained
                    if self.time_contained and not inter.is_time_contained:
                        continue

                    # If requested, skip out-of-time cathode crossers
                    if (
                        self.max_cathode_offset is not None
                        and inter.is_cathode_crosser
                        and abs(inter.cathode_offset) > self.max_cathode_offset
                    ):
                        continue

                    # Fetch the points in the current optical volume
                    sources = self.get_sources(inter)
                    if self.volume == "module":
                        index = self.geo.get_volume_index(sources, volume_id)

                    elif self.volume == "tpc":
                        num_cpm = self.geo.tpc.num_chambers_per_module
                        module_id, tpc_id = volume_id // num_cpm, volume_id % num_cpm
                        index = self.geo.get_volume_index(sources, module_id, tpc_id)
                    else:
                        raise ValueError(f"Volume not recognized: {self.volume}")

                    # If there are no points in this volume, proceed
                    if len(index) == 0:
                        continue

                    # Fetch points and depositions
                    points = self.get_points(inter)[index]
                    depositions = self.get_depositions(inter)[index]
                    if self.ref_volume_id is not None:
                        # If the reference volume is specified, shift positions
                        points = self.geo.translate(
                            points, volume_id, self.ref_volume_id
                        )

                    # Create an interaction which holds positions/depositions
                    inter_v = OutBase(
                        id=inter.id, points=points, depositions=depositions
                    )
                    interactions_v.append(inter_v)

                # Run flash matching
                matches = self.matcher.get_matches(interactions_v, flashes_v)

                # Store flash information
                for inter_v, flash, match in matches:
                    # Get the interaction that matches the cropped version
                    inter = interactions[inter_v.id]

                    # Get the flash hypothesis (if the matcher produces one)
                    hypo_pe, score = -1.0, -1.0
                    if hasattr(match, "hypothesis"):
                        hypo_pe = float(np.sum(list(match.hypothesis)))
                    if hasattr(match, "score"):
                        score = float(match.score)

                    # Update
                    if not inter.is_flash_matched:
                        inter.is_flash_matched = True
                        inter.flash_total_pe = float(flash.total_pe)
                        inter.flash_hypo_pe = hypo_pe
                    else:
                        inter.flash_total_pe += float(flash.total_pe)
                        inter.flash_hypo_pe += hypo_pe

                    if self.merger is not None and not self.update_flashes:
                        orig_flashes = [
                            data[self.flash_key][i] for i in orig_ids[flash.id]
                        ]
                        inter.flash_ids.extend([f.id for f in orig_flashes])
                        inter.flash_volume_ids.extend(
                            [f.volume_id for f in orig_flashes]
                        )
                        inter.flash_times.extend([f.time for f in orig_flashes])
                        inter.flash_scores.extend([score for _ in orig_flashes])
                    else:
                        inter.flash_ids.append(int(flash.id))
                        inter.flash_volume_ids.append(int(flash.volume_id))
                        inter.flash_times.append(float(flash.time))
                        inter.flash_scores.append(score)

            # Cast list attributes to numpy arrays
            for inter in interactions:
                inter.flash_ids = np.asarray(inter.flash_ids, dtype=np.int32)
                inter.flash_volume_ids = np.asarray(
                    inter.flash_volume_ids, dtype=np.int32
                )
                inter.flash_times = np.asarray(inter.flash_times, dtype=np.float32)
                inter.flash_scores = np.asarray(inter.flash_scores, dtype=np.float32)

        # Return an updated flash list, if requested
        if self.update_flashes:
            return {self.flash_key: flashes}



"""Contains Parser class which all parsers inherit from."""

from abc import ABC, abstractmethod


class ParserBase(ABC):
    """Abstract parent class of all parser classes.

    Provides basic functionality shared by all parsers:
    1. Defines a :meth:`__call__` function shared by all classes

    Attributes
    ----------
    name : str
        Name of the parser
    aliases : List[str]
        Aliases of the parser (allowed but disfavored names)
    data_map : dict[str, str]
        Maps function parameter names onto a file data product name
    tree_keys : List[str]
        List of file data product name
    """

    # Name of the parser (as specified in the configuration)
    name = None

    # Alternative allowed names of the parser
    aliases = ()

    # Type of object(s) returned by the parser
    returns = None

    # Overlay method for the objects returned by the parser
    overlay = None

    # List of recognized data type returns
    _data_types = ("tensor", "object", "object_list", "scalar")

    def __init__(self, dtype, **kwargs):
        """Loops over data product names, stores them.

        Parameters
        ----------
        dtype : str
            Data type to cast the input data to
        **kwargs : dict, optional
            Keyword arguments passed to the parser function

        Notes
        -----
        All parser argument which correspond to the name of a tree in the
        LArCV file must be contain either the `_event` or `_event_list` suffix.
        """
        # Store the type in which the parsers should return their data
        self.ftype = dtype
        self.itype = dtype.replace("float", "int")

        # Do a self-consistency check on return data types
        assert self.returns in self._data_types, (
            f"Parser return data type not recognized for '{self.name}': "
            f"{self.returns}. Should be one of {self._data_types}."
        )

        # Find data keys, append them to the map
        self.data_map = {}
        self.tree_keys = []
        for key, value in kwargs.items():
            if "_event" not in key:
                class_name = self.__class__.__name__
                raise TypeError(f"{class_name} got an unexpected argument: {key}.")

            if value is not None:
                self.data_map[key] = value
                if not isinstance(value, list):
                    if value not in self.tree_keys:
                        self.tree_keys.append(value)

                else:
                    for v in value:
                        if v not in self.tree_keys:
                            self.tree_keys.append(v)

    def get_input_data(self, trees):
        """Fetches the required data products from the LArCV data trees, pass
        them to the parser function.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object

        Results
        -------
        object
            Output(s) of the parser function
        """
        # Build the input to the parser function
        data_dict = {}
        for key, value in self.data_map.items():
            if isinstance(value, str):
                if value not in trees:
                    raise ValueError(f"Must provide {value} for parser `{self.name}`.")
                data_dict[key] = trees[value]

            elif isinstance(value, list):
                for v in value:
                    if v not in trees:
                        raise ValueError(f"Must provide {v} for parser `{self.name}`.")
                data_dict[key] = [trees[v] for v in value]

        return data_dict

    @abstractmethod
    def __call__(self, trees):
        """Parse one entry.

        This is a place-holder, must be defined in inheriting class.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        raise NotImplementedError("Must define `__call__` method.")



"""Module that contains all parsers related to LArCV sparse data.

Contains the following parsers:
- :class:`Sparse2DParser`
- :class:`Sparse3DParser`
- :class:`Sparse3DAggregateParser`
- :class:`Sparse3DChargeRescaledParser`
- :class:`Sparse3DGhostParser`
"""

import numpy as np

from spine.data import Meta
from spine.utils.conditional import larcv
from spine.utils.ghost import ChargeRescaler
from spine.utils.globals import GHOST_SHP, SHAPE_PREC

from .base import ParserBase
from .data import ParserTensor

__all__ = [
    "Sparse2DParser",
    "Sparse3DParser",
    "Sparse3DAggregateParser",
    "Sparse3DChargeRescaledParser",
    "Sparse3DGhostParser",
]


class Sparse2DParser(ParserBase):
    """Class that retrieves and parses a 2D sparse tensor.

    .. code-block. yaml

        schema:
          input_data:
            parser: sparse2d
            sparse_event_list:
              - sparse2d_pcluster_0
              - sparse2d_pcluster_1
              - ...
            projection_id: 0
    """

    # Name of the parser (as specified in the configuration)
    name = "parse_sparse2d"

    # Type of object(s) returned by the parser
    returns = "tensor"

    def __init__(self, dtype, projection_id, sparse_event=None, sparse_event_list=None):
        """Initialize the parser.

        Parameters
        ----------
        projection_id : int
            Projection ID to get the 2D images from
        sparse_event: larcv.EventSparseTensor2D, optional
            Sparse tensor to get the voxel/features from
        sparse_event_list: List[larcv.EventSparseTensor2D], optional
            List of sparse tensors to get the voxel/features from
        """
        # Initialize the parent class
        super().__init__(
            dtype, sparse_event=sparse_event, sparse_event_list=sparse_event_list
        )

        # Store the revelant attributes
        self.projection_id = projection_id

        # Get the number of features in the output tensor
        assert (sparse_event is not None) ^ (
            sparse_event_list is not None
        ), "Must provide either `sparse_event` or `sparse_event_list`."
        assert sparse_event_list is None or len(
            sparse_event_list
        ), "Must provide as least 1 sparse_event in the list."

        self.num_features = 1
        if sparse_event_list is not None:
            self.num_features = len(sparse_event_list)

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, sparse_event=None, sparse_event_list=None):
        """Fetches one or a list of tensors, concatenate their feature vectors.

        Parameters
        -------------
        sparse_event: larcv.EventSparseTensor2D, optional
            Sparse tensor to get the voxel/features from
        sparse_event_list: List[larcv.EventSparseTensor2D], optional
            List of sparse tensors to get the voxel/features from

        Returns
        -------
        ParserTensor
            coords : np.ndarray
                (N, 2) array of [x, y] coordinates
            features : np.ndarray
                (N, C) array of [pixel value 0, pixel value 1, ...]
            meta : Meta
                Metadata of the parsed images
        """
        # Parse input into a list
        if sparse_event_list is None:
            sparse_event_list = [sparse_event]

        # Loop over the list of sparse events
        np_voxels, meta, num_points = None, None, None
        np_features = []
        for sparse_event in sparse_event_list:
            # Get the tensor from the appropriate projection
            tensor = sparse_event.sparse_tensor_2d(self.projection_id)

            # Get the shared information
            if meta is None:
                meta = tensor.meta()
                num_points = tensor.as_vector().size()
                np_voxels = np.empty((num_points, 2), dtype=np.int32)
                larcv.fill_2d_voxels(tensor, np_voxels)
                np_voxels = np_voxels.astype(self.itype)
            else:
                assert meta == tensor.meta(), "The metadata must match between tensors."
                assert (
                    num_points == tensor.as_vector().size()
                ), "The number of pixels must match between tensors."

            # Get the feature vector for this tensor
            np_data = np.empty((num_points, 1), dtype=np.float32)
            larcv.fill_2d_pcloud(tensor, np_data)
            np_data = np_data.astype(self.ftype)
            np_features.append(np_data)

        return ParserTensor(
            coords=np_voxels,
            features=np.hstack(np_features),
            meta=Meta.from_larcv(meta),
        )


class Sparse3DParser(ParserBase):
    """Class that retrieves and parses a 3D sparse tensor.

    .. code-block. yaml

        schema:
          input_data:
            parser: sparse3d
            sparse_event_list:
              - sparse3d_pcluster_0
              - sparse3d_pcluster_1
              - ...
    """

    # Name of the parser (as specified in the configuration)
    name = "sparse3d"

    # Type of object(s) returned by the parser
    returns = "tensor"

    def __init__(
        self,
        dtype,
        sparse_event=None,
        sparse_event_list=None,
        num_features=None,
        hit_keys=None,
        nhits_idx=None,
        feature_only=False,
        lexsort=False,
        index_cols=None,
        sum_cols=None,
        prec_col=None,
        precedence=SHAPE_PREC,
    ):
        """Initialize the parser.

        Parameters
        ----------
        sparse_event: larcv.EventSparseTensor3D, optional
            Sparse tensor to get the voxel/features from
        sparse_event_list: List[larcv.EventSparseTensor3D], optional
            List of sparse tensors to get the voxel/features from
        num_features : int, optional
            If a positive integer is specified, the sparse_event_list will be
            split in equal lists of length `features`. Each list will be
            concatenated along the feature dimension separately. Then all
            lists are concatenated along the first dimension (voxels). For
            example, this lets you work with distinct detector volumes whose
            input data is stored in separate TTrees. `num_features` is required
            to be a divider of the `sparse_event_list` length.
        hit_keys : list of int, optional
            Indices among the input features of the `_hit_key_` TTrees that can
            be used to infer the `nhits` quantity (doublet vs triplet point).
        nhits_idx : int, optional
            Index among the input features where the `nhits` feature
            (doublet vs triplet) should be inserted.
        feature_only : bool, default False
            If `True`, only return the feature vector without the coordinates
        lexsort : bool, default False
            When merging points from multiple sources (num_features is not
            `None`), this allows to lexicographically sort coordinates
        index_cols : np.ndarray, optional
            (C) Columns which contain indexes
        sum_cols : np.ndarray, optional
            (S) Columns which should be summed when removing duplicates
        prec_col : int, optional
            Column to be used as a precedence source when removing duplicates
        precedence : np.ndarray, default SHAPE_PREC
            Order of precedence among the classes in prec_col
        """
        # Initialize the parent class
        super().__init__(
            dtype, sparse_event=sparse_event, sparse_event_list=sparse_event_list
        )

        # Store the revelant attributes
        self.num_features = num_features
        self.hit_keys = hit_keys
        self.nhits_idx = nhits_idx
        self.feature_only = feature_only

        # Only lexsort when needed and if there is more than one sparse3d source
        self.lexsort = lexsort
        if self.num_features is None and lexsort:
            raise ValueError(
                "No need to lexsort if there is only one coordinate source."
            )

        # Check on the parameters
        self.compute_nhits = hit_keys is not None
        if self.compute_nhits and nhits_idx is None:
            raise ValueError(
                "The argument nhits_idx needs to be specified if "
                "you want to compute the nhits feature."
            )

        # Get the number of features in the output tensor
        assert (sparse_event is not None) ^ (
            sparse_event_list is not None
        ), "Must provide either `sparse_event` or `sparse_event_list`."
        assert sparse_event_list is None or len(
            sparse_event_list
        ), "Must provide as least 1 sparse_event in the list."

        num_tensors = 1 if sparse_event is not None else len(sparse_event_list)
        if self.num_features is not None:
            if num_tensors % self.num_features != 0:
                raise ValueError(
                    "The `num_features` number in Sparse3DParser should "
                    "be a divider of the `sparse_event_list` length."
                )
        else:
            self.num_features = num_tensors

        # Define the overlay strategy parameters
        self.index_cols = None
        if index_cols is not None:
            self.index_cols = np.asarray(index_cols)
        self.sum_cols = None
        if sum_cols is not None:
            self.sum_cols = np.asarray(sum_cols)
        self.prec_col = prec_col
        self.precedence = precedence

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, sparse_event=None, sparse_event_list=None):
        """Fetches one or a list of tensors, concatenate their feature vectors.

        Parameters
        ----------
        sparse_event: larcv.EventSparseTensor3D, optional
            Sparse tensor to get the voxel/features from
        sparse_event_list: List[larcv.EventSparseTensor3D], optional
            List of sparse tensors to get the voxel/features from

        Returns
        -------
        ParserTensor
            coords : np.ndarray
                (N, 3) array of [x, y, z] coordinates
            features : np.ndarray
                (N, C) array of [pixel value 0, pixel value 1, ...]
            meta : Meta
                Metadata of the parsed images
        """
        # Parse input into a list
        if sparse_event_list is None:
            sparse_event_list = [sparse_event]

        # If requested, split the input list into multiple lists
        split_sparse_event_list = [sparse_event_list]
        if self.num_features != len(sparse_event_list):
            num_groups = len(sparse_event_list) // self.num_features
            split_sparse_event_list = np.split(np.array(sparse_event_list), num_groups)

        # Loop over the individual lists, load the voxels/features
        all_voxels, all_features = [], []
        meta = None
        for sparse_event_list in split_sparse_event_list:
            np_voxels, num_points = None, None
            np_features = []
            hit_key_array = []
            for idx, sparse_event in enumerate(sparse_event_list):
                # Get the shared information
                if meta is None:
                    meta = sparse_event.meta()
                else:
                    assert (
                        meta == sparse_event.meta()
                    ), "The metadata must match between tensors."

                if num_points is None:
                    num_points = sparse_event.as_vector().size()
                    np_voxels = np.empty((num_points, 3), dtype=np.int32)
                    larcv.fill_3d_voxels(sparse_event, np_voxels)
                    np_voxels = np_voxels.astype(self.itype)
                else:
                    assert (
                        num_points == sparse_event.as_vector().size()
                    ), "The number of pixels must match between tensors."

                # Get the feature vector for this tensor
                np_data = np.empty((num_points, 1), dtype=np.float32)
                larcv.fill_3d_pcloud(sparse_event, np_data)
                np_data = np_data.astype(self.ftype)
                np_features.append(np_data)

                # If the number of hits is to be computed, keep track of the
                # required information to do so downstream
                if self.compute_nhits:
                    if idx in self.hit_keys:
                        hit_key_array.append(np_data)

            # If requested, add a feature related to the number of planes
            if self.compute_nhits:
                hit_key_array = np.hstack(hit_key_array)
                nhits = np.sum(hit_key_array >= 0.0, axis=1)[:, -1]
                if self.nhits_idx < 0 or self.nhits_idx > self.num_features:
                    raise ValueError(f"`nhits_idx` ({self.nhits_idx}) is out of range.")
                np_features.insert(self.nhits_idx, nhits)

            # Append to the global list of voxel/features
            all_voxels.append(np_voxels)
            all_features.append(np.hstack(np_features))

        # Stack coordinates/features
        all_voxels = np.vstack(all_voxels)
        all_features = np.vstack(all_features)

        # Lexicographically sort coordinates/features, if requested
        if self.lexsort:
            perm = np.lexsort(all_voxels.T)
            all_voxels = all_voxels[perm]
            all_features = all_features[perm]

        # Return
        return ParserTensor(
            coords=all_voxels,
            features=all_features,
            meta=Meta.from_larcv(meta),
            remove_duplicates=True,
            index_cols=self.index_cols,
            sum_cols=self.sum_cols,
            prec_col=self.prec_col,
            precedence=self.precedence,
            feats_only=self.feature_only,
        )


class Sparse3DAggregateParser(Sparse3DParser):
    """Class that aggregates features from multiple sparse tensors

    .. code-block. yaml

        schema:
          charge_label:
            parser: sparse3d_aggr
            aggr: sum
            sparse_event_list:
              - sparse3d_reco_cryoE_rescaled
              - sparse3d_reco_cryoW_rescaled
    """

    # Name of the parser (as specified in the configuration)
    name = "sparse3d_aggr"

    def __init__(self, dtype, aggr, **kwargs):
        """Initialize the parser.

        Parameters
        ----------
        aggr : str
            Aggregation function to apply ('sum', 'mean', 'max', etc.)
        """
        # Initialize the parent class
        super().__init__(dtype, **kwargs)

        # Store the revelant attributes
        self.aggr_fn = getattr(np, aggr)

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process_aggr(**self.get_input_data(trees))

    def process_aggr(self, sparse_event_list):
        """Fetches a list of tensors, aggregate their feature vectors.

        Parameters
        -------------
        sparse_event_list: List[larcv.EventSparseTensor3D]
            Sparse tensor list to get the voxel/features from

        Returns
        -------
        ParserTensor
            coords : np.ndarray
                (N, 3) array of [x, y, z] coordinates
            features : np.ndarray
                (N, 1) array of aggregated features
            meta : Meta
                Metadata of the parsed image
        """
        # Fetch the list of features using the standard parser
        tensor = self.process(sparse_event_list=sparse_event_list)

        # Combine them into a single feature using the aggregator function
        tensor.features = self.aggr_fn(tensor.features, axis=1)[:, None]

        return tensor


class Sparse3DChargeRescaledParser(Sparse3DParser):
    """Class that convert a tensor containing semantics to binary ghost labels.

    .. code-block. yaml

        schema:
          input_rescaled:
            parser: sparse3d_charge_rescaled
            sparse_event_semantics: sparse3d_semantics
    """

    # Name of the parser (as specified in the configuration)
    name = "parse_sparse3d_rescale_charge"

    # Alternative allowed names of the parser
    aliases = ("parse_sparse3d_charge_rescaled",)

    def __init__(self, dtype, collection_only=False, collection_id=2, **kwargs):
        """Initialize the parser.

        Parameters
        ----------
        collection_only : bool, default False
            If True, only uses the collection plane charge
        collection_id : int, default 2
            Index of the collection plane
        **kwargs : dict, optional
            Data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(dtype, **kwargs)

        # Initialize the charge rescaler
        self.rescaler = ChargeRescaler(collection_only, collection_id)

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process_rescale(**self.get_input_data(trees))

    def process_rescale(self, sparse_event_list):
        """Fetches one or a list of tensors, concatenate their feature vectors.

        Parameters
        -------------
        sparse_event_list: List[larcv.EventSparseTensor3D]
            (7) List of sparse tensors used to compute the rescaled charge
            - Charge value of each of the contributing planes (3)
            - Index of the plane hit contributing to the space point (3)
            - Semantic labels (1)

        Returns
        -------
        ParserTensor
            coords : np.ndarray
                (N, 3) array of [x, y, z] coordinates
            features : np.ndarray
                (N, 1) array of rescaled charge values
            meta : Meta
                Metadata of the parsed image
        """
        # Fetch the list of features using the standard parser
        tensor = self.process(sparse_event_list=sparse_event_list)

        # Use individual hit informations to compute a rescaled charge
        deghost_mask = np.where(tensor.features[:, -1] < GHOST_SHP)[0]
        charges = self.rescaler.process_single(tensor.features[deghost_mask, :-1])

        tensor.features = charges[:, None]

        return tensor


class Sparse3DGhostParser(Sparse3DParser):
    """Class that convert a tensor containing semantics to binary ghost labels.

    .. code-block. yaml

        schema:
          ghost_label:
            parser: sparse3d_ghost
            sparse_event_semantics: sparse3d_semantics
    """

    # Name of the parser (as specified in the configuration)
    name = "sparse3d_ghost"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process_ghost(**self.get_input_data(trees))

    def process_ghost(self, sparse_event):
        """Fetches one or a list of tensors, concatenate their feature vectors.

        Parameters
        -------------
        sparse_event: larcv.EventSparseTensor3D
            Sparse tensor to get the semantic labels

        Returns
        -------
        ParserTensor
            coords : np.ndarray
                (N, 3) array of [x, y, z] coordinates
            features : np.ndarray
                (N, 1) array of ghost labels (1 for ghosts, 0 otherwise)
            meta : Meta
                Metadata of the parsed image
        """
        # Fetch the list of features using the standard parser
        tensor = self.process(sparse_event)

        # Convert the semantics feature to a ghost feature
        tensor.features = (tensor.features == GHOST_SHP).astype(tensor.features.dtype)

        return tensor



"""Module to evaluate diagnostic metrics on showers."""

from spine.ana.base import AnaBase

__all__ = ["ShowerStartDEdxAna"]


class ShowerStartDEdxAna(AnaBase):
    """This analysis script computes the dE/dx value within some distance
    from the start point of an EM shower object.

    This is a useful diagnostic tool to evaluate the calorimetric separability
    of different EM shower types (electron vs photon), which are expected to
    have different dE/dx patterns near their start point.
    """

    # Name of the analysis script (as specified in the configuration)
    name = "shower_start_dedx"

    def __init__(
        self,
        radius,
        obj_type="particle",
        run_mode="both",
        truth_point_mode="points",
        truth_dep_mode="depositions",
        **kwargs,
    ):
        """Initialize the analysis script.

        Parameters
        ----------
        radius : Union[float, List[float]]
            Radius around the start point for which evaluate dE/dx
        **kwargs : dict, optional
            Additional arguments to pass to :class:`AnaBase`
        """
        # Initialize the parent class
        super().__init__(obj_type, run_mode, truth_point_mode, truth_dep_mode, **kwargs)

        # Store parameters
        self.radius = radius

    def process(self, data):
        """Evaluate shower start dE/dx for one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # TODO: Implement shower start dE/dx calculation
        raise NotImplementedError("ShowerStartDEdxAna is not yet implemented.")



"""Numba JIT compiled implementation of basic functions.

Most of these functions are implemented here because vanilla numba does not
support optional arguments, such as `axis` for most functions or
`return_counts` for the `unique` function.
"""

import numba as nb
import numpy as np

__all__ = [
    "seed",
    "unique",
    "sum",
    "mean",
    "mode",
    "argmax",
    "argmin",
    "amax",
    "amin",
    "all",
    "softmax",
    "log_loss",
]


@nb.njit(cache=True)
def seed(seed: nb.int64) -> None:
    """Sets the numpy random seed for all Numba jitted functions.

    Note that setting the seed using `np.random.seed` outside a Numba jitted
    function does *not* set the seed of Numba functions.

    Parameters
    ----------
    seed : int
        Random number generator seed
    """
    np.random.seed(seed)


@nb.njit(cache=True)
def unique(x: nb.int64[:]) -> (nb.int64[:], nb.int64[:]):
    """Numba implementation of `np.unique(x, return_counts=True)`.

    Parameters
    ----------
    x : np.ndarray
        (N) array of values

    Returns
    -------
    np.ndarray
        (U) array of unique values
    np.ndarray
        (U) array of counts of each unique value in the original array
    """
    # Nothing to do if the input is empty
    uniques = np.empty(len(x), dtype=x.dtype)
    counts = np.empty(len(x), dtype=np.int64)
    if len(x) == 0:
        return uniques, counts

    # Build the list of unique values and counts
    x = np.sort(x.flatten())
    uniques[0] = x[0]
    idx = 1
    for i in range(len(x) - 1):
        if x[i] != x[i + 1]:
            uniques[idx] = x[i + 1]
            counts[idx - 1] = i + 1
            idx += 1

    counts[idx - 1] = len(x)

    # Narrow vectors down
    uniques = uniques[:idx]
    counts = counts[:idx]

    # Adjust counts
    counts[1:] = counts[1:] - counts[:-1]

    return uniques, counts


@nb.njit(cache=True)
def sum(x: nb.float32[:, :], axis: nb.int32) -> nb.float32[:]:
    """Numba implementation of `np.sum(x, axis)`.

    Parameters
    ----------
    x : np.ndarray
        (N,M) array of values
    axis : int
        Array axis ID

    Returns
    -------
    np.ndarray
        (N) or (M) array of `sum` values
    """
    assert axis == 0 or axis == 1
    summ = np.empty(x.shape[1 - axis], dtype=x.dtype)
    if axis == 0:
        for i in range(len(summ)):
            summ[i] = np.sum(x[:, i])
    else:
        for i in range(len(summ)):
            summ[i] = np.sum(x[i])

    return summ


@nb.njit(cache=True)
def mean(x: nb.float32[:, :], axis: nb.int32) -> nb.float32[:]:
    """Numba implementation of `np.mean(x, axis)`.

    Parameters
    ----------
    x : np.ndarray
        (N,M) array of values
    axis : int
        Array axis ID

    Returns
    -------
    np.ndarray
        (N) or (M) array of `mean` values
    """
    assert axis == 0 or axis == 1
    mean = np.empty(x.shape[1 - axis], dtype=x.dtype)
    if axis == 0:
        for i in range(len(mean)):
            mean[i] = np.mean(x[:, i])
    else:
        for i in range(len(mean)):
            mean[i] = np.mean(x[i])

    return mean


@nb.njit(cache=True)
def mode(x: nb.int64[:]) -> nb.int64:
    """Numba implementation of `scipy.stats.mode(x)`.

    Parameters
    ----------
    x : np.ndarray
        (N) array of values

    Returns
    -------
    int
        Most-propable value in the array
    """
    values, counts = unique(x)

    return values[np.argmax(counts)]


@nb.njit(cache=True)
def argmin(x: nb.float32[:, :], axis: nb.int32) -> nb.int32[:]:
    """Numba implementation of `np.argmin(x, axis)`.

    Parameters
    ----------
    x : np.ndarray
        (N,M) array of values
    axis : int
        Array axis ID

    Returns
    -------
    np.ndarray
        (N) or (M) array of `argmin` values
    """
    assert axis == 0 or axis == 1
    argmin = np.empty(x.shape[1 - axis], dtype=np.int32)
    if axis == 0:
        for i in range(len(argmin)):
            argmin[i] = np.argmin(x[:, i])
    else:
        for i in range(len(argmin)):
            argmin[i] = np.argmin(x[i])

    return argmin


@nb.njit(cache=True)
def argmax(x: nb.float32[:, :], axis: nb.int32) -> nb.int32[:]:
    """Numba implementation of `np.argmax(x, axis)`.

    Parameters
    ----------
    x : np.ndarray
        (N,M) array of values
    axis : int
        Array axis ID

    Returns
    -------
    np.ndarray
        (N) or (M) array of `argmax` values
    """
    assert axis == 0 or axis == 1
    argmax = np.empty(x.shape[1 - axis], dtype=np.int32)
    if axis == 0:
        for i in range(len(argmax)):
            argmax[i] = np.argmax(x[:, i])

    else:
        for i in range(len(argmax)):
            argmax[i] = np.argmax(x[i])

    return argmax


@nb.njit(cache=True)
def amin(x: nb.float32[:, :], axis: nb.int32) -> nb.float32[:]:
    """Numba implementation of `np.amin(x, axis)`.

    Parameters
    ----------
    x : np.ndarray
        (N,M) array of values
    axis : int
        Array axis ID

    Returns
    -------
    np.ndarray
        (N) or (M) array of `min` values
    """
    assert axis == 0 or axis == 1
    xmin = np.empty(x.shape[1 - axis], dtype=np.int32)
    if axis == 0:
        for i in range(len(xmin)):
            xmin[i] = np.min(x[:, i])

    else:
        for i in range(len(xmin)):
            xmin[i] = np.min(x[i])

    return xmin


@nb.njit(cache=True)
def amax(x: nb.float32[:, :], axis: nb.int32) -> nb.float32[:]:
    """Numba implementation of `np.amax(x, axis)`.

    Parameters
    ----------
    x : np.ndarray
        (N,M) array of values
    axis : int
        Array axis ID

    Returns
    -------
    np.ndarray
        (N) or (M) array of `max` values
    """
    assert axis == 0 or axis == 1
    xmax = np.empty(x.shape[1 - axis], dtype=np.int32)
    if axis == 0:
        for i in range(len(xmax)):
            xmax[i] = np.max(x[:, i])

    else:
        for i in range(len(xmax)):
            xmax[i] = np.max(x[i])

    return xmax


@nb.njit(cache=True)
def all(x: nb.float32[:, :], axis: nb.int32) -> nb.boolean[:]:
    """Numba implementation of `np.all(x, axis)`.

    Parameters
    ----------
    x : np.ndarray
        (N, M) Array of values
    axis : int
        Array axis ID

    Returns
    -------
    np.ndarray
        (N) or (M) array of `all` outputs
    """
    assert axis == 0 or axis == 1
    all = np.empty(x.shape[1 - axis], dtype=np.bool_)
    if axis == 0:
        for i in range(len(all)):
            all[i] = np.all(x[:, i])

    else:
        for i in range(len(all)):
            all[i] = np.all(x[i])

    return all


@nb.njit(cache=True)
def softmax(x: nb.float32[:, :], axis: nb.int32) -> nb.float32[:, :]:
    """
    Numba implementation of `scipy.special.softmax(x, axis)`.

    Parameters
    ----------
    x : np.ndarray
        (N,M) array of values
    axis : int
        Array axis ID

    Returns
    -------
    np.ndarray
        (N,M) Array of softmax scores
    """
    assert axis == 0 or axis == 1
    if axis == 0:
        xmax = amax(x, axis=0)
        logsumexp = np.log(np.sum(np.exp(x - xmax), axis=0)) + xmax
        return np.exp(x - logsumexp)
    else:
        xmax = amax(x, axis=1).reshape(-1, 1)
        logsumexp = np.log(np.sum(np.exp(x - xmax), axis=1)).reshape(-1, 1) + xmax
        return np.exp(x - logsumexp)


@nb.njit(cache=True)
def log_loss(label: nb.boolean[:], pred: nb.float32[:]) -> nb.float32:
    """Numba implementation of cross-entropy loss.

    Parameters
    ----------
    label : np.ndarray
        (N) array of boolean labels (0 or 1)
    pred : np.ndarray
        (N) array of float scores (between 0 and 1)

    Returns
    -------
    float
        Cross-entropy loss
    """
    if len(label) > 0:
        return -(
            np.sum(np.log(pred[label])) + np.sum(np.log(1.0 - pred[~label]))
        ) / len(label)
    else:
        return 0.0



import MinkowskiEngine as ME
import torch
import torch.nn as nn
from scipy.special import logit

from .act_norm import act_factory, norm_factory
from .blocks import ConvolutionBlock, ResNetBlock
from .configuration import setup_cnn_configuration


class SparseGenerator(torch.nn.Module):

    def __init__(self, cfg, name="sparse_generator"):
        super(SparseGenerator, self).__init__()
        setup_cnn_configuration(self, cfg, name)

        self.model_config = cfg[name]
        self.reps = self.model_config.get("reps", 2)
        self.depth = self.model_config.get("depth", 7)
        self.num_filters = self.model_config.get("num_filters", 16)
        # self.nPlanes = [(2**i) * self.num_filters for i in range(self.depth)]
        self.nPlanes = [16, 16, 32, 32, 64, 64, 128, 128, 256]
        print(self.nPlanes)
        assert len(self.nPlanes) == self.depth
        self.latent_size = self.model_config.get("latent_size", 512)
        final_tensor_shape = self.spatial_size // (2 ** (self.depth - 1))
        self.coordConv = self.model_config.get("coordConv", False)
        print("Final Tensor Shape = ", final_tensor_shape)
        self.resolution = self.model_config.get("resolution", 1024)
        self.threshold = logit(self.model_config.get("threshold", 0.0))
        print(self.threshold)
        self.layer_limit = self.model_config.get("layer_limit", -1)
        if self.layer_limit < 0:
            self.layer_limit = len(self.nPlanes) + 1

        self.linear = nn.Sequential(
            norm_factory(self.norm, self.latent_size, **self.norm_args),
            act_factory(self.activation_name, **self.activation_args),
            ME.MinkowskiConvolutionTranspose(
                in_channels=self.latent_size,
                out_channels=self.nPlanes[-1],
                kernel_size=final_tensor_shape,
                stride=final_tensor_shape,
                dimension=self.D,
                bias=self.allow_bias,
                generate_new_coords=True,
            ),
        )
        # Initialize Decoder
        self.decoding_block = []
        self.decoding_conv = []
        self.layer_cls = []
        for i in range(self.depth - 2, -1, -1):
            m = []
            m.append(norm_factory(self.norm, self.nPlanes[i + 1], **self.norm_args))
            m.append(act_factory(self.activation_name, **self.activation_args))
            m.append(
                ME.MinkowskiConvolutionTranspose(
                    in_channels=self.nPlanes[i + 1],
                    out_channels=self.nPlanes[i],
                    kernel_size=2,
                    stride=2,
                    dimension=self.D,
                    bias=self.allow_bias,
                    generate_new_coords=True,
                )
            )
            m = nn.Sequential(*m)
            self.decoding_conv.append(m)
            m = []
            for j in range(self.reps):
                m.append(
                    ResNetBlock(
                        self.nPlanes[i],
                        self.nPlanes[i],
                        dimension=self.D,
                        activation=self.activation_name,
                        activation_args=self.activation_args,
                        normalization=self.norm,
                        normalization_args=self.norm_args,
                        has_bias=self.allow_bias,
                    )
                )
            m = nn.Sequential(*m)
            self.decoding_block.append(m)
            self.layer_cls.append(
                ME.MinkowskiLinear(self.nPlanes[i], 1, bias=self.allow_bias)
            )
        self.decoding_block = nn.Sequential(*self.decoding_block)
        self.decoding_conv = nn.Sequential(*self.decoding_conv)
        self.layer_cls = nn.Sequential(*self.layer_cls)

        # pruning
        self.pruning = ME.MinkowskiPruning()

    def get_batch_indices(self, out):
        return out.coords_man.get_row_indices_per_batch(out.coords_key)

    def get_target(self, out, target_key, kernel_size=1):
        with torch.no_grad():
            target = torch.zeros(len(out), dtype=torch.bool)
            cm = out.coords_man
            strided_target_key = cm.stride(
                target_key, out.tensor_stride[0], force_creation=True
            )
            ins, outs = cm.get_kernel_map(
                out.coords_key,
                strided_target_key,
                kernel_size=kernel_size,
                region_type=1,
            )
            for curr_in in ins:
                target[curr_in] = 1
        return target

    def forward(self, latent, target_key):

        out_cls, targets = [], []

        latent.set_tensor_stride(self.resolution)

        x = self.linear(latent)
        layer_count = 0

        for i, layer in enumerate(self.decoding_conv):
            print(layer_count)
            if layer_count >= self.layer_limit:
                break
            x = layer(x)
            x = self.decoding_block[i](x)
            x_cls = self.layer_cls[i](x)
            target = self.get_target(x, target_key)
            targets.append(target)
            out_cls.append(x_cls)
            layer_count += 1
            keep = (x_cls.F > self.threshold).cpu().squeeze()
            if self.training:
                keep += target
            if keep.sum() > 0:
                x = self.pruning(x, keep.cpu())
            else:
                break

        return {"reconstruction": x, "out_cls": out_cls, "targets": targets}


class SparseGeneratorSimple(torch.nn.Module):

    def __init__(self, cfg, name="sparse_generator"):
        super(SparseGeneratorSimple, self).__init__()
        setup_cnn_configuration(self, cfg, name)

        self.model_config = cfg[name]
        self.reps = self.model_config.get("reps", 2)
        self.depth = self.model_config.get("depth", 7)
        self.num_filters = self.model_config.get("num_filters", 16)
        # self.nPlanes = [(2**i) * self.num_filters for i in range(self.depth)]
        self.nPlanes = [16, 16, 32, 32, 64, 64, 128, 128, 256]
        print(self.nPlanes)
        assert len(self.nPlanes) == self.depth
        self.latent_size = self.model_config.get("latent_size", 512)
        final_tensor_shape = self.spatial_size // (2 ** (self.depth - 1))
        self.coordConv = self.model_config.get("coordConv", False)
        print("Final Tensor Shape = ", final_tensor_shape)
        self.resolution = self.model_config.get("resolution", 1024)
        self.threshold = logit(self.model_config.get("threshold", 0.0))
        print(self.threshold)
        self.layer_limit = self.model_config.get("layer_limit", -1)
        if self.layer_limit < 0:
            self.layer_limit = len(self.nPlanes) + 1

        self.linear = nn.Sequential(
            norm_factory(self.norm, self.latent_size, **self.norm_args),
            act_factory(self.activation_name, **self.activation_args),
            ME.MinkowskiConvolutionTranspose(
                in_channels=self.latent_size,
                out_channels=self.nPlanes[-1],
                kernel_size=final_tensor_shape,
                stride=final_tensor_shape,
                dimension=self.D,
                bias=self.allow_bias,
                generate_new_coords=True,
            ),
        )
        # Initialize Decoder
        self.decoding_block = []
        self.decoding_conv = []
        self.layer_cls = []
        for i in range(self.depth - 2, -1, -1):
            m = []
            m.append(norm_factory(self.norm, self.nPlanes[i + 1], **self.norm_args))
            m.append(act_factory(self.activation_name, **self.activation_args))
            m.append(
                ME.MinkowskiConvolutionTranspose(
                    in_channels=self.nPlanes[i + 1],
                    out_channels=self.nPlanes[i],
                    kernel_size=2,
                    stride=2,
                    dimension=self.D,
                    bias=self.allow_bias,
                    generate_new_coords=True,
                )
            )
            m = nn.Sequential(*m)
            self.decoding_conv.append(m)
            m = []
            for j in range(self.reps):
                m.append(
                    ConvolutionBlock(
                        self.nPlanes[i],
                        self.nPlanes[i],
                        dimension=self.D,
                        activation=self.activation_name,
                        activation_args=self.activation_args,
                        normalization=self.norm,
                        normalization_args=self.norm_args,
                        has_bias=self.allow_bias,
                    )
                )
            m = nn.Sequential(*m)
            self.decoding_block.append(m)
            self.layer_cls.append(
                ME.MinkowskiLinear(self.nPlanes[i], 1, bias=self.allow_bias)
            )
        self.decoding_block = nn.Sequential(*self.decoding_block)
        self.decoding_conv = nn.Sequential(*self.decoding_conv)
        self.layer_cls = nn.Sequential(*self.layer_cls)

        # pruning
        self.pruning = ME.MinkowskiPruning()

    def get_batch_indices(self, out):
        return out.coords_man.get_row_indices_per_batch(out.coords_key)

    def get_target(self, out, target_key, kernel_size=1):
        with torch.no_grad():
            target = torch.zeros(len(out), dtype=torch.bool)
            cm = out.coords_man
            strided_target_key = cm.stride(
                target_key, out.tensor_stride[0], force_creation=True
            )
            ins, outs = cm.get_kernel_map(
                out.coords_key,
                strided_target_key,
                kernel_size=kernel_size,
                region_type=1,
            )
            for curr_in in ins:
                target[curr_in] = 1
        return target

    def forward(self, latent, target_key):

        out_cls, targets = [], []

        latent.set_tensor_stride(self.resolution)

        x = self.linear(latent)
        layer_count = 0

        for i, layer in enumerate(self.decoding_conv):
            print(layer_count)
            if layer_count >= self.layer_limit:
                break
            x = layer(x)
            x = self.decoding_block[i](x)
            x_cls = self.layer_cls[i](x)
            target = self.get_target(x, target_key)
            targets.append(target)
            out_cls.append(x_cls)
            layer_count += 1
            keep = (x_cls.F > self.threshold).cpu().squeeze()
            if self.training:
                keep += target
            if keep.sum() > 0:
                x = self.pruning(x, keep.cpu())
            else:
                break

        return {"reconstruction": x, "out_cls": out_cls, "targets": targets}



import torch
import torch.nn as nn
from torch_geometric.nn import MLP, PointNetConv, fps, global_max_pool, radius

# From Pytorch Geometric Examples for PointNet:
# https://github.com/pyg-team/pytorch_geometric/blob/master/examples/pointnet2_classification.py


class SAModule(torch.nn.Module):
    def __init__(self, ratio, r, nn):
        super().__init__()
        self.ratio = ratio
        self.r = r
        self.conv = PointNetConv(nn, add_self_loops=False)

    def forward(self, x, pos, batch):
        idx = fps(pos, batch, ratio=self.ratio)
        row, col = radius(
            pos, pos[idx], self.r, batch, batch[idx], max_num_neighbors=64
        )
        edge_index = torch.stack([col, row], dim=0)
        x_dst = None if x is None else x[idx]
        x = self.conv((x, x_dst), (pos, pos[idx]), edge_index)
        pos, batch = pos[idx], batch[idx]
        return x, pos, batch


class GlobalSAModule(torch.nn.Module):
    def __init__(self, nn):
        super().__init__()
        self.nn = nn

    def forward(self, x, pos, batch):
        x = self.nn(torch.cat([x, pos], dim=1))
        x = global_max_pool(x, batch)
        pos = pos.new_zeros((x.size(0), 3))
        batch = torch.arange(x.size(0), device=batch.device)
        return x, pos, batch


class PointNet(torch.nn.Module):
    """
    Pytorch Geometric's implementation of PointNet, modified for
    use in lartpc_spine3d and generalized.
    """

    def __init__(self, cfg, name="pointnet"):
        super(PointNet, self).__init__()

        self.model_config = cfg[name]

        self.depth = self.model_config.get("depth", 2)

        self.sampling_ratio = self.model_config.get("sampling_ratio", 0.5)
        if isinstance(self.sampling_ratio, float):
            self.sampling_ratio = [self.sampling_ratio] * self.depth
        elif isinstance(self.sampling_ratio, list):
            assert len(self.sampling_ratio) == self.depth
        else:
            raise ValueError("Sampling ratio must either be given as \
                             float or list of floats.")

        self.neighbor_radius = self.model_config.get("neighbor_radius", 3.0)
        if isinstance(self.neighbor_radius, float):
            self.neighbor_radius = [self.neighbor_radius] * self.depth
        elif isinstance(self.neighbor_radius, list):
            assert len(self.neighbor_radius) == self.depth
        else:
            raise ValueError("Neighbor aggregation radius must either \
                             be given as float or list of floats.")

        self.mlp_specs = []
        self.sa_modules = nn.ModuleList()

        for i in range(self.depth):
            mlp_specs = self.model_config["mlp_specs_{}".format(i)]
            self.sa_modules.append(
                SAModule(
                    self.sampling_ratio[i], self.neighbor_radius[i], MLP(mlp_specs)
                )
            )
            self.mlp_specs.append(mlp_specs)

        self.mlp_specs_glob = self.model_config.get(
            "mlp_specs_glob", [256 + 3, 256, 512, 1024]
        )
        self.mlp_specs_final = self.model_config.get(
            "mlp_specs_final", [1024, 512, 256, 128]
        )
        self.dropout = self.model_config.get("dropout", 0.5)
        self.latent_size = self.mlp_specs_final[-1]

        self.sa3_module = GlobalSAModule(MLP(self.mlp_specs_glob))
        self.mlp = MLP(self.mlp_specs_final, dropout=self.dropout, norm=None)

    def forward(self, data):
        sa0_out = (data.x, data.pos, data.batch)

        out = sa0_out

        for m in self.sa_modules:
            out = m(*out)

        sa3_out = self.sa3_module(*out)
        x, pos, batch = sa3_out

        return self.mlp(x)


class PointNetEncoder(torch.nn.Module):

    def __init__(self, cfg, name="pointnet_encoder"):
        super(PointNetEncoder, self).__init__()
        self.net = PointNet(cfg)
        self.latent_size = self.net.latent_size

    def forward(self, batch):
        out = self.net(batch)
        return out



from collections import defaultdict
from pprint import pprint

import torch
import torch.nn as nn

from .misc import *


class SPICELoss(nn.Module):
    """
    Loss function for Sparse Spatial Embeddings Model, with fixed
    centroids and symmetric gaussian kernels.
    """

    def __init__(self, cfg, name="spice_loss"):
        super(SPICELoss, self).__init__()
        self.loss_config = cfg[name]
        self.seediness_weight = self.loss_config.get("seediness_weight", 1.0)
        self.embedding_weight = self.loss_config.get("embedding_weight", 1.0)
        self.smoothing_weight = self.loss_config.get("smoothing_weight", 1.0)
        self.spatial_size = self.loss_config.get("spatial_size", 768)

        self.mask_loss_fn = self.loss_config.get("mask_loss_fn", "BCE")
        self.seed_loss_fn = self.loss_config.get("seed_loss_fn", "L1")
        self.batch_loc = self.loss_config.get("batch_loc", 0)

        if self.mask_loss_fn == "BCE":
            self.mask_loss = nn.BCEWithLogitsLoss(reduction="none")
        elif self.mask_loss_fn == "lovasz_hinge":
            self.mask_loss = LovaszHingeLoss()
        elif self.mask_loss_fn == "focal":
            raise NotImplementedError
        else:
            raise ValueError("Invalid loss scheme: {}".format(self.loss_scheme))

        # L2 Loss for Seediness
        if self.seed_loss_fn == "L1":
            self.seed_loss = torch.nn.L1Loss(reduction="mean")
        elif self.seed_loss_fn == "L1":
            self.seed_loss = torch.nn.MSELoss(reduction="mean")
        elif self.seed_loss_fn == "huber":
            self.seed_loss = torch.nn.SmoothL1Loss(reduction="mean")
        else:
            raise ValueError("Invalid loss scheme: {}".format(self.loss_scheme))

    def find_cluster_means(self, features, labels):
        """
        For a given image, compute the centroids mu_c for each
        cluster label in the embedding space.
        Inputs:
            features (torch.Tensor): the pixel embeddings, shape=(N, d) where
            N is the number of pixels and d is the embedding space dimension.
            labels (torch.Tensor): ground-truth group labels, shape=(N, )
        Returns:
            cluster_means (torch.Tensor): (n_c, d) tensor where n_c is the number of
            distinct instances. Each row is a (1,d) vector corresponding to
            the coordinates of the i-th centroid.
        """
        cluster_means = find_cluster_means(features, labels)
        return cluster_means

    def get_per_class_probabilities(self, embeddings, margins, labels, eps=1e-6):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        n = labels.shape[0]
        centroids = self.find_cluster_means(embeddings, labels)
        sigma = scatter_mean(margins.squeeze(), labels)
        num_clusters = labels.unique().shape[0]

        # Compute spatial term
        em = embeddings[:, None, :]
        centroids = centroids[None, :, :]
        sqdists = ((em - centroids) ** 2).sum(-1)

        p = sqdists / (2.0 * sigma.view(1, -1) ** 2)
        p = torch.clamp(torch.exp(-p), min=eps, max=1 - eps)
        logits = torch.logit(p, eps=eps)
        eye = torch.eye(len(labels.unique()), dtype=torch.float32, device=device)
        targets = eye[labels]
        loss_tensor = self.mask_loss(logits, targets)
        loss = loss_tensor.mean(dim=0).mean()
        with torch.no_grad():
            acc = iou_batch(logits > 0, targets.bool())
        smoothing_loss = margin_smoothing_loss(
            margins.squeeze(), sigma.detach(), labels, margin=0
        )
        p = torch.gather(p, 1, labels.view(-1, 1))
        return loss, smoothing_loss, p.squeeze(), acc

    def combine_multiclass(
        self, embeddings, margins, seediness, slabels, clabels, coords
    ):
        """
        Wrapper function for combining different components of the loss,
        in particular when clustering must be done PER SEMANTIC CLASS.

        NOTE: When there are multiple semantic classes, we compute the DLoss
        by first masking out by each semantic segmentation (ground-truth/prediction)
        and then compute the clustering loss over each masked point cloud.

        INPUTS:
            features (torch.Tensor): pixel embeddings
            slabels (torch.Tensor): semantic labels
            clabels (torch.Tensor): group/instance/cluster labels
        OUTPUT:
            loss_segs (list): list of computed loss values for each semantic class.
            loss[i] = computed DLoss for semantic class <i>.
            acc_segs (list): list of computed clustering accuracy for each semantic class.
        """
        loss = defaultdict(list)
        accuracy = defaultdict(float)
        semantic_classes = slabels.unique()
        for sc in semantic_classes:
            if int(sc) == 4:  # Skip low energy deposits
                continue
            index = slabels == sc
            clabels_unique, _ = unique_label_torch(clabels[index])
            mask_loss, smoothing_loss, probs, acc = self.get_per_class_probabilities(
                embeddings[index], margins[index], clabels_unique
            )
            prob_truth = probs.detach()
            seed_loss = self.seed_loss(prob_truth, seediness[index].squeeze(1))
            total_loss = (
                self.embedding_weight * mask_loss
                + self.seediness_weight * seed_loss
                + self.smoothing_weight * smoothing_loss
            )
            loss["loss"].append(total_loss)
            loss["mask_loss"].append(float(self.embedding_weight * mask_loss))
            loss["seed_loss"].append(float(self.seediness_weight * seed_loss))
            loss["smoothing_loss"].append(float(self.smoothing_weight * smoothing_loss))
            loss["mask_loss_{}".format(int(sc))].append(float(mask_loss))
            loss["seed_loss_{}".format(int(sc))].append(float(seed_loss))
            accuracy["accuracy_{}".format(int(sc))] = acc

        return loss, accuracy

    def forward(self, out, segment_label, group_label):

        num_gpus = len(segment_label)
        loss = defaultdict(list)
        accuracy = defaultdict(list)

        for i in range(num_gpus):
            slabels = segment_label[i][:, -1]
            # coords = segment_label[i][:, :3].float()
            # if torch.cuda.is_available():
            #    coords = coords.cuda()
            slabels = slabels.int()
            clabels = group_label[i][:, -1]
            batch_idx = segment_label[i][:, self.batch_loc]
            embedding = out["embeddings"][i]
            seediness = out["seediness"][i]
            margins = out["margins"][i]
            nbatch = batch_idx.unique().shape[0]

            for bidx in batch_idx.unique(sorted=True):
                embedding_batch = embedding[batch_idx == bidx]
                slabels_batch = slabels[batch_idx == bidx]
                clabels_batch = clabels[batch_idx == bidx]
                seed_batch = seediness[batch_idx == bidx]
                margins_batch = margins[batch_idx == bidx]

                loss_class, acc_class = self.combine_multiclass(
                    embedding_batch,
                    margins_batch,
                    seed_batch,
                    slabels_batch,
                    clabels_batch,
                )
                if len(acc_class.values()):
                    for key, val in loss_class.items():
                        loss[key].append(sum(val) / len(val))
                    for s, acc in acc_class.items():
                        accuracy[s].append(acc)
                    acc = sum(acc_class.values()) / len(acc_class.values())
                    accuracy["accuracy"].append(acc)

        loss_avg = {}
        acc_avg = defaultdict(float)

        for key, val in loss.items():
            loss_avg[key] = sum(val) / len(val)
        for key, val in accuracy.items():
            acc_avg[key] = sum(val) / len(val)

        res = {}
        res.update(loss_avg)
        res.update(acc_avg)

        return res


class SPICEInterLoss(SPICELoss):

    def __init__(self, cfg, name="spice_loss"):
        super(SPICEInterLoss, self).__init__(cfg, name)
        self.inter_weight = self.loss_config.get("inter_weight", 1.0)
        self.inter_margin = self.loss_config.get("inter_margin", 0.2)
        self.norm = 2
        self._min_voxels = self.loss_config.get("min_voxels", 2)

    def regularization(self, cluster_means):
        """
        Implementation of regularization loss in Discriminative Loss
        Inputs:
            cluster_means (torch.Tensor): output from find_cluster_means
        Returns:
            reg_loss (float): computed regularization loss (see paper).
        """
        reg_loss = regularization_loss(cluster_means)
        return reg_loss

    def inter_cluster_loss(self, cluster_means, margin=0.2):
        """
        Implementation of distance loss in Discriminative Loss.
        Inputs:
            cluster_means (torch.Tensor): output from find_cluster_means
            margin (float/int): the magnitude of the margin delta_d in the paper.
            Think of it as the distance between each separate clusters in
            embedding space.
        Returns:
            inter_loss (float): computed cross-centroid distance loss (see paper).
            Factor of 2 is included for proper normalization.
        """
        inter_loss = inter_cluster_loss(cluster_means, margin=margin)
        return inter_loss

    def get_per_class_probabilities(self, embeddings, margins, labels, eps=1e-6):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        n = labels.shape[0]
        centroids = self.find_cluster_means(embeddings, labels)
        sigma = self.find_cluster_means(margins, labels).view(-1, 1)
        smoothing_loss = margin_smoothing_loss(
            margins.view(-1), sigma.view(-1).detach(), labels, margin=0
        )

        num_clusters = labels.unique().shape[0]
        inter_loss = self.inter_cluster_loss(centroids, margin=self.inter_margin)

        # Compute spatial term
        em = embeddings[:, None, :]
        centroids = centroids[None, :, :]
        cov = torch.clamp(sigma[:, 0][None, :], min=eps)
        sqdists = ((em - centroids) ** 2).sum(-1) / (2.0 * cov**2)

        pvec = torch.exp(-sqdists)
        logits = torch.logit(pvec, eps=eps)
        # print(logits)
        eye = torch.eye(len(labels.unique()), dtype=torch.float32, device=device)
        targets = eye[labels]
        loss = self.mask_loss(logits, targets).mean()
        # loss = loss_tensor
        with torch.no_grad():
            acc = iou_batch(logits > 0, targets.bool())
        p = torch.gather(pvec, 1, labels.view(-1, 1))
        loss += inter_loss
        return loss, smoothing_loss, float(inter_loss), p.squeeze(), acc

    def combine_multiclass(self, embeddings, margins, seediness, slabels, clabels):
        """
        Wrapper function for combining different components of the loss,
        in particular when clustering must be done PER SEMANTIC CLASS.

        NOTE: When there are multiple semantic classes, we compute the DLoss
        by first masking out by each semantic segmentation (ground-truth/prediction)
        and then compute the clustering loss over each masked point cloud.

        INPUTS:
            features (torch.Tensor): pixel embeddings
            slabels (torch.Tensor): semantic labels
            clabels (torch.Tensor): group/instance/cluster labels
        OUTPUT:
            loss_segs (list): list of computed loss values for each semantic class.
            loss[i] = computed DLoss for semantic class <i>.
            acc_segs (list): list of computed clustering accuracy for each semantic class.
        """
        loss = defaultdict(list)
        accuracy = defaultdict(float)
        semantic_classes = slabels.unique()
        for sc in semantic_classes:
            if int(sc) == 4:
                continue
            index = slabels == sc
            if len(embeddings[index]) < self._min_voxels:
                continue
            clabels_unique, _ = unique_label_torch(clabels[index])
            mask_loss, smoothing_loss, inter_loss, probs, acc = (
                self.get_per_class_probabilities(
                    embeddings[index], margins[index], clabels_unique
                )
            )
            prob_truth = probs.detach()
            seed_loss = self.seed_loss(prob_truth, seediness[index].squeeze(1))
            total_loss = (
                self.embedding_weight * mask_loss
                + self.seediness_weight * seed_loss
                + self.smoothing_weight * smoothing_loss
            )
            loss["loss"].append(total_loss)
            loss["mask_loss"].append(float(self.embedding_weight * mask_loss))
            loss["seed_loss"].append(float(self.seediness_weight * seed_loss))
            loss["smoothing_loss"].append(float(self.smoothing_weight * smoothing_loss))
            loss["inter_loss"].append(float(self.inter_weight * inter_loss))
            loss["mask_loss_{}".format(int(sc))].append(float(mask_loss))
            loss["seed_loss_{}".format(int(sc))].append(float(seed_loss))
            accuracy["accuracy_{}".format(int(sc))] = acc

        return loss, accuracy


class SPICEAttractorLoss(SPICEInterLoss):

    def __init__(self, cfg, name="spice_loss"):
        super(SPICEAttractorLoss, self).__init__(cfg, name)
        self.inter_weight = self.loss_config.get("inter_weight", 1.0)
        self.inter_margin = self.loss_config.get("inter_margin", 0.2)
        self.norm = 2
        self._min_voxels = self.loss_config.get("min_voxels", 2)

        self.n_attractors = 10
        self.min_points_for_attractors = 100
        self.attractor_margin = 0.0
        self.attractor_weight = 1.0

    def sample_attractors(self, embeddings, labels, centroids):

        attractor_loss = 0.0
        attractor_intra_loss = 0.0
        attractor_inter_loss = 0.0

        frag_labels, counts = labels.unique(return_counts=True)
        # Unique label for attractors
        emb_to_att = -torch.ones(embeddings.shape[0], device=embeddings.device).long()
        # Embeddings of attractors and fragment labels of attractors
        attractors, attractor_labels = [], []
        masked_centroids = []
        inds_start = 0
        for frag, count in zip(frag_labels, counts):
            em = embeddings[labels == frag]
            if count >= self.min_points_for_attractors:
                perm = torch.randperm(count)[: self.n_attractors]
                att = em[perm]
                dist = torch.cdist(em, att)
                _, inds = torch.min(dist, dim=1)
                inds += inds_start
                attractors.append(att)
                attractor_labels.append(
                    frag * torch.ones(att.shape[0], device=embeddings.device).long()
                )
                masked_centroids.append(centroids[frag])
                emb_to_att[labels == frag] = inds
                inds_start += self.n_attractors

        if len(attractors) == 0:
            return 0.0

        attractors = torch.vstack(attractors)
        attractor_labels = torch.cat(attractor_labels, dim=0)

        masked_centroids = torch.vstack(masked_centroids)

        # Compute embedding to attractor loss
        masked_embeddings = embeddings[emb_to_att != -1]
        emb_to_att = emb_to_att[emb_to_att != -1]

        # Avoid zero placeholders by relabeling from 0 to number of attractors
        emb_to_att, _ = unique_label_torch(emb_to_att)
        attractor_intra_loss = intra_cluster_loss(
            masked_embeddings, attractors, emb_to_att, margin=self.attractor_margin
        )
        attractor_labels, _ = unique_label_torch(attractor_labels)
        attractor_inter_loss = intra_cluster_loss(
            attractors, masked_centroids, attractor_labels, margin=self.attractor_margin
        )

        print(attractor_intra_loss, attractor_inter_loss)

        attractor_loss = attractor_intra_loss + attractor_inter_loss

        # Compute attractor to centroid loss
        # print(attractor_loss)
        return attractor_loss

    def get_per_class_probabilities(self, embeddings, margins, labels, eps=1e-6):
        """
        Computes binary foreground/background loss.
        """
        device = embeddings.device
        n = labels.shape[0]
        centroids = self.find_cluster_means(embeddings, labels)
        sigma = self.find_cluster_means(margins, labels).view(-1, 1)
        smoothing_loss = margin_smoothing_loss(
            margins.squeeze(), sigma.view(-1).detach(), labels, margin=0
        )

        num_clusters = labels.unique().shape[0]
        inter_loss = self.inter_cluster_loss(centroids, margin=self.inter_margin)

        # Compute attractor loss
        attractor_loss = self.sample_attractors(embeddings, labels, centroids)

        # Compute spatial term
        em = embeddings[:, None, :]
        centroids = centroids[None, :, :]
        cov = torch.clamp(sigma[:, 0][None, :], min=eps)
        sqdists = ((em - centroids) ** 2).sum(-1) / (2.0 * cov**2)

        pvec = torch.exp(-sqdists)
        logits = torch.logit(pvec, eps=eps)
        # print(logits)
        eye = torch.eye(len(labels.unique()), dtype=torch.float32, device=device)
        targets = eye[labels]
        loss = self.mask_loss(logits, targets).mean()
        # loss = loss_tensor
        with torch.no_grad():
            acc = iou_batch(logits > 0, targets.bool())
        p = torch.gather(pvec, 1, labels.view(-1, 1))
        loss += self.inter_weight * inter_loss
        loss += self.attractor_weight * attractor_loss
        return (
            loss,
            smoothing_loss,
            float(inter_loss),
            float(attractor_loss),
            p.squeeze(),
            acc,
        )

    def combine_multiclass(self, embeddings, margins, seediness, slabels, clabels):
        """
        Wrapper function for combining different components of the loss,
        in particular when clustering must be done PER SEMANTIC CLASS.

        NOTE: When there are multiple semantic classes, we compute the DLoss
        by first masking out by each semantic segmentation (ground-truth/prediction)
        and then compute the clustering loss over each masked point cloud.

        INPUTS:
            features (torch.Tensor): pixel embeddings
            slabels (torch.Tensor): semantic labels
            clabels (torch.Tensor): group/instance/cluster labels
        OUTPUT:
            loss_segs (list): list of computed loss values for each semantic class.
            loss[i] = computed DLoss for semantic class <i>.
            acc_segs (list): list of computed clustering accuracy for each semantic class.
        """
        loss = defaultdict(list)
        accuracy = defaultdict(float)
        semantic_classes = slabels.unique()
        for sc in semantic_classes:
            if int(sc) == 4:
                continue
            index = slabels == sc
            if len(embeddings[index]) < self._min_voxels:
                continue
            clabels_unique, _ = unique_label_torch(clabels[index])
            embedding_loss = self.get_per_class_probabilities(
                embeddings[index], margins[index], clabels_unique
            )

            mask_loss = embedding_loss[0]
            smoothing_loss = embedding_loss[1]
            inter_loss = embedding_loss[2]
            attractor_loss = embedding_loss[3]
            probs = embedding_loss[4]
            acc = embedding_loss[5]

            prob_truth = probs.detach()
            seed_loss = self.seed_loss(prob_truth, seediness[index].squeeze(1))
            total_loss = (
                self.embedding_weight * mask_loss
                + self.seediness_weight * seed_loss
                + self.smoothing_weight * smoothing_loss
            )
            loss["loss"].append(total_loss)
            loss["mask_loss"].append(float(self.embedding_weight * mask_loss))
            loss["seed_loss"].append(float(self.seediness_weight * seed_loss))
            loss["smoothing_loss"].append(float(self.smoothing_weight * smoothing_loss))
            loss["inter_loss"].append(float(self.inter_weight * inter_loss))
            loss["attractor_loss"].append(float(attractor_loss))
            loss["mask_loss_{}".format(int(sc))].append(float(mask_loss))
            loss["seed_loss_{}".format(int(sc))].append(float(seed_loss))
            accuracy["accuracy_{}".format(int(sc))] = acc

        return loss, accuracy
